/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

//#include <backends/sstmacro-native/manager-env.h>
#include <sstmac/backends/native/manager.h>
#include <sstmac/backends/native/nodelist.h>

#include <sstmac/common/driver_util.h>
#include <sstmac/backends/native/eventmap.h>

#include <sstmac/common/counter.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/cast.h>
#include <sstmac/software/ami/environment.h>

#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>

#include <iostream>
#include <iterator>
#include <cstdlib>

#include <sstmac/common/factories/node_factory.h>
#include <sstmac/common/factories/nic_factory.h>

namespace sstmac
{
  namespace native
  {
    using namespace std;
    using namespace sstmac::sw;
    using namespace sstmac::hw;

    //
    // Default constructor.
    //
    manager::manager(const eventmanager::ptr& ev_mgr) :
      eventmanager_(ev_mgr), nodes_(nodelist::construct()), running_(false),
          next_ppid_(0), dbg_("<debug> manager")
    //stats(statscollector::construct("stats.log"))
    {
      SSTMAC_DEBUG << "manager::manager()\n";
      eventmanager::global = eventmanager_;
      logger::timer_ = eventmanager_;
    }

    //
    // Construct a manager.
    //
    manager::ptr
    manager::construct(const eventmanager::ptr& ev_mgr)
    {
      manager::ptr retval(new manager(ev_mgr));
      return retval;
    }

    //
    // Goodbye.
    //
    manager::~manager() throw ()
    {
      SSTMAC_DEBUG << "manager::~manager()\n";
      if (this->running_)
      {
          std::cerr
              << "FATAL:  manager going out of scope while still running.\n";
          abort();
      }
       SSTMAC_DEBUG << "manager::~manager() leaving\n";
    }

    //
    // Define a network.
    //
    void
    manager::set_interconnect(const interconnect::ptr &net)
    {
      SSTMAC_DEBUG << "manager::set_interconnect(" << net->to_string() << ")\n";
      if (running_)
        {
          throw illformederror(
              "manager::set_interconnect: Simulation is running.");
        }
      interconnect_ = net;

      //stats->add_stats(interconnect_->get_stats());

    }

    //
    // Access the interconnect.
    //
    interconnect::const_ptr
    manager::get_interconnect() const
    {
      return interconnect_;
    }

    //
    // Number of nodes currently defined.
    //
    int64_t
    manager::nodes() const
    {

      return nodes_->size();
    }

    //
    // Number of open network slots.
    //
    int64_t
    manager::open_slots() const
    {
      if (interconnect_ == 0)
        {
          throw nullerror("manager::open_slots:  null interconnect pointer.");
        }
      if (interconnect_->slots() <= 0)
        {
          return -1;
        }
      else
        {
          return (interconnect_->slots() - this->nodes());
        }
    }

    node::ptr
    manager::get_node(nodeid::ptr id) const
    {
      return nodes_->get(id);
    }

    //
    // Add one or more copies of the given node type to the manager.
    //
    long
    manager::add_nodes(sim_parameters::ptr params)
    {
      SSTMAC_DEBUG << "manager::add_nodes() \n";

      std::string failmodel = params->get_optional_param("failure_model",
          "none");
      if (!fail_ && failmodel != "none")
        {
          SSTMAC_DEBUG
                << "manager::add_nodes() -- constructing failure model \n";
          fail_ = failure_config::construct(params);
        }

      sstmac::hw::node::ptr type = SSTNodeFactory::get_node(params->get_param("node_name"), params);
      sstmac::hw::networkinterface::ptr nictype = SSTNicFactory::get_nic(params->get_param("nic_name"), params,
          type, interconnect_);

      if (interconnect_ == 0)
        {
          throw nullerror("manager::add_nodes:  No interconnect defined yet.");
        }
      if (type == 0)
        {
          throw nullerror("manager::add_nodes:  Null node pointer\n");
        }

      long count = interconnect_->slots();
      if (count != interconnect_->slots())
        {

          std::cout << "count: " << count << "\n";
          std::cout << "interconnect size: " << interconnect_->slots() << "\n";
          throw illformederror(
              "node_num parameter does not match interconnect size. "
                "node_num should never be specified in the input file.");
        }

      SSTMAC_DEBUG << "manager::add_nodes(" << type << ", " << count << ")\n";

      /** DEPRECATED CHECK: No more node_num parameter
       if ((interconnect_->slots() > 0)
       && (count > int(interconnect_->slots() - this->nodes())))
       {
       throw valueerror("manager::add_nodes:  Insufficient slots available.");
       }
       */
      if (count < 0)
        {
          std::cerr << "manager::add_nodes:  Simulator has unlimited size. "
              << " Unlimited interconnect size is no longer valid."
              << std::endl;
          throw illformederror(
              "manager::add_nodes:  Infinite interconnect size.");

          /** DEPRECATED node_num parameter
           if (interconnect_->slots() <= 0)
           {
           std::cerr << "manager::add_nodes:  Simulator has unlimited size, "
           << "so an unlimited node count is not possible\n";
           throw illformederror("manager::add_nodes:  Need infinite memory.");
           }
           count = check_cast<int>(interconnect_->slots()) - this->nodes();
           */
        }
      for (long i = 0; i < count; ++i)
        {
          nodeaddress::ptr id = nodeaddress::convert_from_uniqueid(i);
          node::ptr newnode = type->clone(id);

          sstmac::hw::networkinterface::ptr newnic = nictype->clone(newnode,
              interconnect_, params);
          newnode->set_nic(newnic);

          nodes_->add(newnode);
          SSTMAC_DEBUG << "manager: adding nodeid " << id->to_string() << "\n";
          //available_.insert(id);
        }

      //finish off the template node
      type->finish();
      nictype->finish();

      return this->open_slots();
    }

    //
    // Setup the simulation.
    //
    void
    manager::setup()
    {
      SSTMAC_DEBUG << "manager::setup: entered\n";
      if (interconnect_ == 0)
        {
          throw nullerror("manager::setup:  Null interconnect pointer.");
        }
      if (running_)
        {
          throw illformederror("manager::setup:  Simulation already running.");
        }
      // Initial setup.
      if (eventmanager_ == 0)
        {
          eventmanager_ = eventmap::construct();
          eventmanager::global = eventmanager_;
        }
      // Make sure the network simulator knows about our event manager.
      interconnect_->set_eventmanager(eventmanager_);

      vis::vis_display::ptr visd =
          boost::dynamic_pointer_cast<vis::vis_display>(interconnect_);
      if (visd && sstmac_env::params->has_param("vis_engine"))
        {
          visd->vis_start(true);
          std::cout
              << "--- VTK visualization is on, press any key to continue \n";
          std::cin.ignore();
        }

      // Initialize the nodes.
      for (int i = 0; i < nodes_->size(); ++i)
        {
          nodeid::ptr id = nodeid::construct(i);
          node::ptr the_node = nodes_->get(id);
          the_node->initialize(eventmanager_);
          the_node->get_nic()->set_eventmanager(eventmanager_);

          //the_node->add_networkinterface(simplenic::construct(the_node,
          //                                                    interconnect_));
        }

      if (fail_)
        {
          fail_->config_nodes(nodes_, eventmanager_);

          fail_->config_network(interconnect_);
        }

      // Schedule job startups.
      // for (startup_t::iterator it = startup_.begin(); it != startup_.end(); ++it)
      //  {
      //    eventmanager_->schedule(it->first, it->second, sst_message::ptr());
      //  }
      startup_.clear();
      running_ = true;
    }

    //
    // Start the simulation.
    //
    timestamp
    manager::run(timestamp until)
    {

      // setup the simulation before starting
      setup();

      SSTMAC_DEBUG << "manager::run: calling eventmanager::run\n";

      if (until.sec() > 0)
        {
          eventmanager_->schedule(until, stophandler::construct(),
              sst_message::ptr());
        }

      eventmanager_->run();

      // Now call done routine to end simulation and print Stats.
      done();

      return eventmanager_->now();

    }

    void
    manager::done()
    {
      SSTMAC_DEBUG << "manager::done: entered\n";

      running_ = false;
      /*if (allocated_.size() > 0)
       {
       std::cout << "manager no longer running but " << allocated_.size()
       << " nodes are still allocated--deallocating all nodes."
       << std::endl;
       std::copy(allocated_.begin(), allocated_.end(),
       std::inserter(available_, available_.begin()));
       allocated_.clear();
       }*/

      vis::vis_display::ptr visd =
          boost::dynamic_pointer_cast<vis::vis_display>(interconnect_);
      if (visd && sstmac_env::params->has_param("vis_engine"))
        {
          std::cout << "--- Simulation has finished, press any key to kill \n";
          std::cin.ignore();
          visd->vis_complete();
        }

      interconnect_->finish();
      for (int i=0; i < nodes_->size(); ++i){
          nodeid::ptr id = nodeid::construct(i);
          node::ptr the_node = nodes_->get(id);
          the_node->finish();
      }

      // stats->printStats(eventmanager_->now());
    }

  }
} // end of namespace sstmac
