/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/common/driver_util.h>

#include <sstmac/software/process/taskid.h>

#include <sstmac/common/config.h>
#include <sstmac/common/fileio.h>



#include <sstmac/common/basicstringtokenizer.h>




#ifdef HAVE_VTK
#include <sstmac/common/vis/vtk_engine.h>
#endif

namespace sstmac
{

  using namespace std;
  using namespace sstmac::hw;
  using namespace sstmac::sw;
  using namespace boost;

  void
  file_parameters::parse_file(std::string fname)
  {
    std::ifstream in;
    SSTFileIO::open_file(in, fname);

    if (in.is_open())
      {

        std::string line;

        while (in.good())
          {
            std::getline(in, line);

            if (line[0] == '#')
              { //a comment
                continue;
              }
            else if (line.find("=") != std::string::npos)
              { //an assignment
                pair<string, string> keyval = sim_parameters::parse_param(line);

                add_param_override(keyval.first, keyval.second);
              }
            else if (line.find("include") != string::npos)
              { //an include line
                std::string f = line.substr(7);
                std::string dir = "";
                size_t pos = fname.find_last_of('/');
                if (pos != string::npos)
                  {
                    dir = fname.substr(0, pos + 1);
                  }
                f = trim_str(f);
                parse_file(dir + f);
              }
          }

      }
    else
      {
        std::stringstream ss;
        ss << "file_parameters: Failed to open input file \"" << fname << "\""
            << std::endl;

        throw valueerror(ss.str());
      }
  }

  file_parameters::file_parameters(std::string fname)
  {
    parse_file(fname);

  }






  sstmac::vis::vis_engine::ptr
  construct_vis_engine(sstmac::sim_parameters::ptr params)
  {
    vis::vis_engine::ptr veng;
    if (params->has_param("vis_engine"))
      {
#ifdef HAVE_VTK
        if (params->get_param("vis_engine") == "vtk")
          {
            veng = vis::vtk_engine::construct();
          }
        else
          {
            throw ssterror("unknown vis engine specified");
          }
#endif
      }

    return veng;
  }

  double
  get_positive_float(const char *value, bool &errorflag)
  {
    char *endptr = NULL;
    double val = strtod(value, &endptr);
    // This is really pretty bad form
    errno = 0;
    if (*value == '\0' || *endptr != '\0')
      {
        errorflag = true;
        std::cerr << "Not a valid floating point value: " << value << "\n";
      }
    if (errno == ERANGE)
      {
        errorflag = true;
        std::cerr << "Value outside double precision range: " << value << "\n";
      }
    if (val <= 0)
      {
        errorflag = true;
        std::cerr << "Not a positive floating point value: " << value << "\n";
      }
    return val;
  }

  int
  get_int(const char *value, bool &errorflag)
  {
    char *endptr;
    errno = 0;
    long val = strtol(value, &endptr, 10);
    if (*value == '\0' || *endptr != '\0')
      {
        errorflag = true;
        std::cerr << "Invalid integer value: " << value << "\n";
      }
    if (errno == ERANGE || val > std::numeric_limits<int>::max() || val
        < std::numeric_limits<int>::min())
      {
        errorflag = true;
        std::cerr << "Value outside 32-bit integer range: " << value << "\n";
      }
    if (errno == EINVAL)
      {
        errorflag = true;
        std::cerr << "Value contains invalid characters: " << value << "\n";
      }
    return val;
  }

  int
  get_positive_int(const char *value, bool &errorflag)
  {
    int v = get_int(value, errorflag);
    if ((!errorflag) && v <= 0)
      {
        errorflag = true;
        std::cerr << "Not a positive non-zero integer value: " << value << "\n";
      }
    return v;
  }

  std::vector<int>
  get_intvec(const char *value, bool &errorflag)
  {
    std::vector<int> retval;
    char *brkt;
    char *val = strdup(value);
    const char *sep = " ,;_-";
    for (char *word = strtok_r(val, sep, &brkt); word != NULL; word = strtok_r(
        NULL, sep, &brkt))
      {
        retval.push_back(get_int(word, errorflag));
        if (errorflag)
          break;
      }
    free(val);
    return retval;
  }

  std::string
  file_parameters::get_param(std::string key) const
  {
    SSTMAC_DEBUG << "file_parameters: getting param " << key << "\n";
    maptype::const_iterator it = params_.find(key);
    if (it == params_.end())
      {
        //std::cout << "parameter named: " << key << std::endl;
        //throw sstmac::valueerror("file_parameters: could not find parameter");
        sst_throw(sstmac::valueerror, "file_parameters: could not find parameter %s", key.c_str());
    }
    return it->second;
  }

  command_parameters::command_parameters()
  {
  }

    command_parameters::ptr
    command_parameters::construct()
    {
      return ptr(new command_parameters());
    }

    std::string
    command_parameters::get_param(std::string key) const
    {

      maptype::const_iterator it = params_.find(key);
      if (it == params_.end())
        {
          std::cout << "parameter named: " << key << std::endl;
          throw sstmac::valueerror(
              "native_parameters: could not find parameter");
        }
      return it->second;
    }

    bool
    command_parameters::has_param(std::string key) const
    {
      maptype::const_iterator it = params_.find(key);
      return (it != params_.end());
    }

    void
    command_parameters::add_param(std::string key, std::string val)
    {
      SSTMAC_DEBUG << "setting key " << key << " to value " << val << "\n";

      if (params_.find(key) != params_.end())
        {
          throw sstmac::valueerror(
              "command_parameters::add_param - key already in params");
        }
      else
        {
          params_[key] = val;
        }
    }

    void
    command_parameters::add_param_override(std::string key, std::string val)
    {
      SSTMAC_DEBUG << "setting key " << key << " to value " << val << "...";

      if (params_.find(key) != params_.end())
        {
          SSTMAC_DEBUG << "overriding \n";
        }
      else
        {
          SSTMAC_DEBUG << "adding new param \n";
        }

      params_[key] = val;

    }

    void
    command_parameters::combine_into(sim_parameters::ptr sp)
    {
      BOOST_FOREACH(maptype::value_type it, params_)
              {
                sp->add_param_override(it.first, it.second);
              }
    }

    std::string
    command_parameters::to_string() const
    {
      return "command_parameters";
    }

    command_parameters::~command_parameters()
    {
    }

    file_parameters::ptr
    file_parameters::construct(std::string filename)
    {
      return ptr(new file_parameters(filename));
    }

    void
    file_parameters::append_params(std::string filename)
    {
      parse_file(filename);
    }

    bool
    file_parameters::has_param(std::string key) const
    {
      maptype::const_iterator it = params_.find(key);
      return (it != params_.end());
    }

    void
    file_parameters::add_param(std::string key, std::string val)
    {
      SSTMAC_DEBUG << "setting key " << key << " to value " << val << "\n";

      if (params_.find(key) != params_.end())
        {
          throw sstmac::valueerror(
              "file_parameters::add_param - key already in params");
        }
      else
        {
          params_[key] = val;
        }
    }

    void
    file_parameters::add_param_override(std::string key, std::string val)
    {
      SSTMAC_DEBUG << "setting key " << key << " to value " << val << "...";

      if (params_.find(key) != params_.end())
        {
          SSTMAC_DEBUG << "overriding \n";
        }
      else
        {
          SSTMAC_DEBUG << "adding new param \n";
        }

      params_[key] = val;

    }

    void
    file_parameters::combine_into(sim_parameters::ptr sp)
    {
      BOOST_FOREACH(maptype::value_type it, params_)
              {
                sp->add_param_override(it.first, it.second);
              }
    }

    std::string
    file_parameters::to_string() const
    {
      return "file_parameters";
    }

    file_parameters::~file_parameters()
    {
        //there seems to be a memory leak in the boost implementation
        //but explicit clearing seems to fix it
        params_.clear();
    }


}
