/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_EVENTCALLBACK_H_INCLUDED
#define SSTMAC_COMMON_EVENTCALLBACK_H_INCLUDED

#include <sstmac/common/eventhandler.h>
#include <sstmac/common/errors.h>

namespace sstmac
{

  /**
   * The main interface for something that can respond to an event (sst_message).
   */
  class eventcallback : public eventhandler
  {
  public:
    typedef boost::intrusive_ptr<eventcallback> ptr;

    timestamp time_;

  protected:
    eventcallback()
        : time_(0)
    {
    }

  public:
    virtual ~eventcallback()
    {
    }

    virtual void
    handle(const sst_message::ptr& msg)
    {
        callback(msg);
    }

    virtual void
    callback(const sst_message::ptr& msg) = 0;

    virtual std::string
    to_string() const
    {
      return "event callback";
    }

    timestamp
    get_time() const
    {
      return time_;
    }

    void
    set_time(const timestamp& t)
    {
      time_ = t;
    }

  };

  /** This throws an exception because messages with callbacks should never be handled */
  class callbackhandler : public eventhandler
  {
    protected:
     callbackhandler()
     {
     }

    public:
      typedef boost::intrusive_ptr<callbackhandler> ptr;

      static ptr
      construct()
      {
          return new callbackhandler();
      }

      void handle(const sst_message::ptr &msg)
      {
          if (!msg->has_callback()){
              sst_throw(illformederror, "callback handler received msg without callback");
          }
          eventhandler::ptr cb = msg->pop_callback();
          cb->handle(msg);
      }
  };

  template<class Cls, typename Fxn>
    class eventcallback_msg : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

    public:
      eventcallback_msg(const boost::intrusive_ptr<Cls>& obj, Fxn fxn) :
        obj_(obj), fxn_(fxn)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        Cls& obj = *obj_;
        (obj_.get()->*fxn_)(msg);
      }
    };

  template<class Cls, typename Fxn, typename Arg1, typename Arg2>
    class eventcallback_msg_2_args : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

      Arg1 arg1_;

      Arg2 arg2_;

    public:
      eventcallback_msg_2_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn,
          const Arg1& arg1, const Arg2& arg2) :
        obj_(obj), fxn_(fxn), arg1_(arg1), arg2_(arg2)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        Cls& obj = *obj_;
        (obj_.get()->*fxn_)(msg, arg1_, arg2_);
      }
    };

  template<class Cls, typename Fxn>
    eventcallback::ptr
    msg_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn)
    {
      eventcallback::ptr callback = new eventcallback_msg<Cls, Fxn> (cls, fxn);
      return callback;
    }

  template<class Cls, typename Fxn, typename Arg1, typename Arg2>
    eventcallback::ptr
    msg_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1, const Arg2& arg2)
    {
      eventcallback::ptr callback = new eventcallback_msg_2_args<Cls, Fxn,
          Arg1, Arg2> (cls, fxn, arg1, arg2);
      return callback;
    }

  template<class Cls, typename Fxn>
    class eventcallback_tmpl_0_args : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

    public:
      eventcallback_tmpl_0_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn) :
        obj_(obj), fxn_(fxn)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        Cls& obj = *obj_;
        (obj_.get()->*fxn_)();
      }

    };

  template<class Cls, typename Fxn, typename Arg1>
    class eventcallback_tmpl_1_args : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

      Arg1 arg1_;

    public:
      eventcallback_tmpl_1_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn,
          const Arg1& arg1) :
        obj_(obj), fxn_(fxn), arg1_(arg1)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        Cls& obj = *obj_;
        (obj_.get()->*fxn_)(arg1_);
      }

    };

  template<class Cls, typename Fxn, typename Arg1, typename Arg2>
    class eventcallback_tmpl_2_args : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

      Arg1 arg1_;

      Arg2 arg2_;

    public:
      eventcallback_tmpl_2_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn,
          const Arg1& arg1, const Arg2& arg2) :
        obj_(obj), fxn_(fxn), arg1_(arg1), arg2_(arg2)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        Cls& obj = *obj_;
        (obj_.get()->*fxn_)(arg1_, arg2_);
      }

    };

  template<class Cls, typename Fxn, typename Arg1, typename Arg2, typename Arg3>
    class eventcallback_tmpl_3_args : public eventcallback
    {
    private:
      boost::intrusive_ptr<Cls> obj_;

      Fxn fxn_;

      Arg1 arg1_;

      Arg2 arg2_;

      Arg3 arg3_;

    public:
      eventcallback_tmpl_3_args(const boost::intrusive_ptr<Cls>& obj, Fxn fxn,
          const Arg1& arg1, const Arg2& arg2, const Arg3& arg3) :
        obj_(obj), fxn_(fxn), arg1_(arg1), arg2_(arg2), arg3_(arg3)
      {
      }

      void
      callback(const sst_message::ptr& msg)
      {
        (obj_.get()->*fxn_)(arg1_, arg2_, arg3_);
      }

    };

  template<class Cls, typename Fxn>
    eventcallback::ptr
    new_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_0_args<Cls, Fxn> (
          cls, fxn);
      return callback;
    }

  template<class Cls, typename Fxn, typename Arg1>
    eventcallback::ptr
    new_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_1_args<Cls, Fxn,
          Arg1> (cls, fxn, arg1);
      return callback;
    }

  template<class Cls, typename Fxn, typename Arg1, typename Arg2>
    eventcallback::ptr
    new_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1, const Arg2& arg2)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_2_args<Cls, Fxn,
          Arg1, Arg2> (cls, fxn, arg1, arg2);
      return callback;
    }

  template<class Cls, typename Fxn, typename Arg1, typename Arg2, typename Arg3>
    eventcallback::ptr
    new_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1, const Arg2& arg2, const Arg3& arg3)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_3_args<Cls, Fxn,
          Arg1, Arg2, Arg3> (cls, fxn, arg1, arg2, arg3);
      return callback;
    }

  template<class Cls, typename Fxn, typename Arg1>
    sst_message::ptr
    new_message_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_1_args<Cls, Fxn,
          Arg1> (cls, fxn, arg1);
      sst_message::ptr msg = sst_message::construct();
      msg->add_callback(callback);
      return msg;
    }

  template<class Cls, typename Fxn, typename Arg1, typename Arg2>
     sst_message::ptr
     new_message_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
         const Arg1& arg1, const Arg2& arg2)
     {
       eventcallback::ptr callback = new eventcallback_tmpl_2_args<Cls, Fxn,
           Arg1, Arg2> (cls, fxn, arg1, arg2);
       sst_message::ptr msg = sst_message::construct();
       msg->add_callback(callback);
       return msg;
     }

  template<class Cls, typename Fxn, typename Arg1, typename Arg2, typename Arg3>
    sst_message::ptr
    new_message_callback(const boost::intrusive_ptr<Cls>& cls, Fxn fxn,
        const Arg1& arg1, const Arg2& arg2, const Arg3& arg3)
    {
      eventcallback::ptr callback = new eventcallback_tmpl_3_args<Cls, Fxn,
          Arg1, Arg2, Arg3> (cls, fxn, arg1, arg2, arg3);
      sst_message::ptr msg = sst_message::construct();
      msg->add_callback(callback);
      return msg;
    }

} // end of namespace sstmac
#endif
