/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/common/factories/launch_factory.h>
#include <sstmac/common/factories/allocator_factory.h>
#include <sstmac/common/factories/indexer_factory.h>

#include <unistd.h>
#include <getopt.h>

namespace sstmac
{
  using namespace sstmac::sw;
  using namespace std;

  std::map<std::string, SSTDesc_base*>* SSTLaunchFactory::Launchmap_ = 0;

  void
  SSTLaunchFactory::register_Launch(std::string name, SSTDesc_base* descr)
  {
    if (Launchmap_ == 0)
      {
        Launchmap_ = new std::map<std::string, SSTDesc_base*>();
      }

    add_to_map(name, descr, Launchmap_);
  }

  void
  SSTLaunchFactory::parse_aprun(const std::string& cmd, long& nproc, long& nproc_per_node,
      std::vector<int>& core_affinities, std::string& appname)
  {
    int aprun_numa_containment = 0;
    option aprun_gopt[] =
      {
        { "n", required_argument, NULL, 'n' },
        { "N", required_argument, NULL, 'N' },
        { "ss", no_argument, &aprun_numa_containment, 1 },
        { "cc", required_argument, NULL, 'C' },
        { "S", required_argument, NULL, 'S' },
        { "d", required_argument, NULL, 'd' },
        { NULL, 0, NULL, '\0' } };

    char cmdline_str[200];
    ::strcpy(cmdline_str, cmd.c_str());
    char *argv[50];
    int argc = 0;
    char* pch = strtok(cmdline_str, " ");
    while (pch != 0)
      {
        argv[argc] = pch;
        ++argc;
        pch = strtok(0, " ");
      }

    /** The last argument is the name of the actual app */
    appname = argv[argc - 1];
    --argc;

    int _nproc = -1;
    int _nproc_per_node = -1;
    int ch;
    int option_index = 1;
    std::string core_aff_str;
    optind = 1;
    while ((ch = getopt_long_only(argc, argv, "n:N:S:d:", aprun_gopt,
        &option_index)) != -1)
      {
        switch (ch)
          {
        case 0: //this set an input flag
          break;
        case 'n':
          _nproc = atoi(optarg);
          break;
        case 'N':
          _nproc_per_node = atoi(optarg);
          break;
        case 'S':
          break;
        case 'C':
          core_aff_str = optarg;
          break;
        default:
          break;
          }
      }
    if (_nproc <= 0)
      throw inputerror(
          "aprun allocator did not receive a valid -n specification");

    if (_nproc_per_node == 0 || _nproc_per_node > _nproc)
      throw inputerror(
          "aprun allocator did not receive a valid -N specification");

    if (core_aff_str.size() > 0) //we got a core affinity spec
      {
        string tosep = core_aff_str;
        std::deque<std::string> tok;
        std::string space = ",";
        pst::BasicStringTokenizer::tokenize(tosep, tok, space);
        std::deque<std::string>::const_iterator it, end = tok.end();
        for (it = tok.begin(); it != end; ++it)
          {
            std::string core = *it;
            core_affinities.push_back(atoi(core.c_str()));
          }
      }

    nproc = _nproc;
    nproc_per_node = _nproc_per_node;
  }

  std::list<launchinfo::ptr>
  SSTLaunchFactory::construct_launchers(sim_parameters::ptr params)
  {
    std::list<launchinfo::ptr> ret;

    int appnum = 1;
    std::stringstream ss;
    while (true)
      {
        ss << "launch_app" << appnum;
        if (!params->has_param(ss.str()))
          {
            break;
          }

        std::string launch_prefix = ss.str();
        std::string launch_param = params->get_param(launch_prefix);
        sstmac::sw::appid aid(appnum);
        timestamp start(0);

        /** Check for an aprun launch */
        std::string launchcmd = "";
        size_t pos = launch_param.find_first_of(' ');
        if (pos != string::npos)
          {
            launchcmd = launch_param.substr(0, pos);
          }

        long tasks = 0;
        long ntask_per_node = 0;
        std::string appname;
        std::vector<int> core_affinities;
        if (launchcmd == "aprun") //parse the aprun options
          {
            parse_aprun(launch_param, tasks, ntask_per_node, core_affinities,
                appname);
            if (ntask_per_node == -1) //nothing given
              {
                int ncores = params->get_optional_int_param("node_cores", 1);
                ntask_per_node = ncores > tasks ? tasks : ncores;
              }
          }
        else //standard launch
          {
            ss << "_start";
            start = params->get_time_param(ss.str());
            tasks
                = params->get_optional_long_param(launch_prefix + "_size", -1);
            ntask_per_node = params->get_optional_long_param(
                launch_prefix + "_ntask_per_node", 1);
            appname = launch_param;

            if (params->has_param(launch_prefix + "_core_affinities"))
              {
                params->get_vector_param(launch_prefix + "_core_affinities",
                    core_affinities);
              }
          }

        long extra = params->get_optional_long_param("launch_extra_nodes", 0);
        launchinfo::ptr li = launchinfo::construct(appname, aid, params, start,
            tasks, ntask_per_node, extra, core_affinities);
        ret.push_back(li);

        appnum++;
        ss.str("");
      }

    return ret;
  }

  sstmac::sw::launcher::ptr
  SSTLaunchFactory::get_Launch(std::string name, sim_parameters::ptr params,
      sstmac::hw::topology::ptr top)
  {

    if (Launchmap_->find(name) == Launchmap_->end())
      {
        std::cout << "Launch name: " << name << "\n";
        throw ssterror(
            "SSTAppFactory::get_Launch -- Launch was not found in Launch map, did you forget to call SSTRegisterLaunch?");
      }

    ptr_type::ptr p = (*Launchmap_)[name]->build(params);
    launcher::ptr cast = boost::dynamic_pointer_cast<launcher>(p);
    allocationstrategy::ptr alloc = SSTAllocatorFactory::get_allocator(
        params->get_param("launch_allocation"), params, top);
    indexstrategy::ptr ind = SSTIndexerFactory::get_Indexer(
        params->get_param("launch_indexing"), params);
    std::list<launchinfo::ptr> linfos = construct_launchers(params);
    cast->init(linfos, alloc, ind);

    return cast;
  }

}
