/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/common/messages/sst_message.h>
#include <sstmac/common/eventcallback.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/messages/sst_metadata.h>
#include <sstmac/software/process/operatingsystem.h>

namespace sstmac
{

    sst_message::sst_message(MESSAGE_TYPES t) :
        kind_(0), msgtype_(t), needs_ack_(true), is_scheduled_(false)
    {
    }

  sst_message::sst_message(std::string lib, timestamp t,
      sst_message::MESSAGE_TYPES ty, int k) :
    lib_name_(lib), time_(t), msgtype_(ty), kind_(k), needs_ack_(true), is_scheduled_(false)
  {

  }

  sst_message::sst_message(std::string libn, timestamp t,
      nodeaddress::ptr from, nodeaddress::ptr to,
      sst_message::MESSAGE_TYPES ty, int k) :
    lib_name_(libn), time_(t), msgtype_(ty), fromaddr_(from), toaddr_(to),
        kind_(k), needs_ack_(true), is_scheduled_(false)
  {

  }

  sst_message::~sst_message()
  {
    // std::cout << "calling sst_message destructor now" << std::endl;
  }

  //
  // Allocate an object of this type.
  //

  sst_message::ptr
  sst_message::construct(std::string lib, timestamp t, MESSAGE_TYPES ty, int k)
  {
    sst_message::ptr retval(new sst_message(lib, t, ty, k));

    return retval;
  }

  sst_message::ptr
  sst_message::construct(std::string lib, timestamp t, nodeaddress::ptr from,
      nodeaddress::ptr to, MESSAGE_TYPES ty, int k)
  {
    sst_message::ptr retval(new sst_message(lib, t, from, to, ty, k));

    return retval;
  }

  void
  sst_message::serialize(sst_serializer* ser)
  {
    // std::cout << "in sst_message serialize" << std::endl;
    ser->pack(lib_name_);
    ser->pack(time_);
    fromaddr_->serialize(ser);
    toaddr_->serialize(ser);

    //std::cout << "completed sst_message serialize" << std::endl;
  }
  sst_message::ptr
  sst_message::deserialize(sst_serializer* ser)
  {
    sst_message::ptr ret = sst_message::construct();
    ret->do_deser(ser);

    return ret;

  }

  const sw::taskid&
  sst_message::dest_task() const
  {
      throw illformederror("generic sst message does not have dest task");
  }

  const sw::taskid&
  sst_message::source_task() const
  {
      throw illformederror("generic sst message does not have src task");
  }

  long
  sst_message::source_thread(const sw::operatingsystem::ptr& os) const
  {
      throw illformederror("generic sst message does not have source thread");
  }

  long
  sst_message::dest_thread(const sw::operatingsystem::ptr& os) const
  {
      throw illformederror("generic sst message does not have dest thread");
  }

  void
  sst_message::do_deser(sst_serializer* ser)
  {
    ser->unpack(lib_name_);
    ser->unpack(time_);
    fromaddr_ = ser->get_deser_addr();
    fromaddr_->deserialize(ser);
    toaddr_ = ser->get_deser_addr();
    toaddr_->deserialize(ser);
  }

  bool
  sst_message::has_callback() const
  {
      return !callbacks_.empty();
  }

  void
  sst_message::add_callback(const eventhandler::ptr& callback)
  {
      callbacks_.push_front(callback);
  }

  eventhandler::ptr
  sst_message::pop_callback()
  {
      eventhandler::ptr tmp = callbacks_.front();
      callbacks_.pop_front();
      return tmp;
  }

  sst_message::ptr
  sst_message::clone(MESSAGE_TYPES t) const
  {
      return construct(lib_name_, time_, fromaddr_, toaddr_, t, kind_);
  }

  sst_message::ptr
  sst_message::clone() const
  {
      return construct(lib_name_, time_, fromaddr_, toaddr_, msgtype_, kind_);
  }

  void
  sst_message::mutate(mutation_t mut)
  {
      throw unimplementederror("message cannot be mutated");
  }

  long
  sst_message::get_byte_length() const
  {
      return 0;
  }

  const timestamp&
  sst_message::get_event_time() const
  {
      throw unimplementederror("sst_message does not implement get_event_time interface");
  }

  void
  sst_message::set_event_time(const timestamp &t)
  {
      throw unimplementederror("sst_message does not implement set_event_time interface");
  }

  int
  sst_message::get_core() const
  {
      throw unimplementederror("sst_message does not implement get_core interface");
  }

  void
  sst_message::set_core(int core)
  {
      throw unimplementederror("sst_message does not implement set_core interface");
  }

  sw::AMI::AMI_COMP_FUNC
  sst_message::get_function() const
  {
      throw unimplementederror("sst_message does not implement get_function interface");
  }

  bool
  sst_message::has_event_value(const std::string& key) const
  {
      throw unimplementederror("sst_message does not implement has_event_value interface");
  }

  unsigned long long
  sst_message::get_event_value(const std::string& key) const
  {
      throw unimplementederror("sst_message does not implement get_event_value interface");
  }

  unsigned long long
  sst_message::get_optional_event_value(const std::string& key) const
  {
      throw unimplementederror("sst_message does not implement get_optional_event_value interface");
  }

  void
  sst_message::set_event_value(const std::string& key, unsigned long long val)
  {
      throw unimplementederror("sst_message does not implement set_event_value interface");
  }

#define casestr(x) case x: return #x
  const char*
  sst_message::tostr(MESSAGE_TYPES t)
  {
      switch(t)
      {
          casestr(SST);
          casestr(MPI);
          casestr(ACTOR);
          casestr(SEND_ACK);
          casestr(COMPUTE);
          casestr(LAUNCH);
          casestr(DISK);
          casestr(FILEMETA);
          casestr(FILEDATA);
          casestr(HPX);
          casestr(FAILURE);
          casestr(SDMA);
          casestr(NETWORK);
          casestr(RDMA_GET_REQ);
          casestr(RDMA_PUT_REQ);
          casestr(RDMA_GET_RSP);
          casestr(RDMA_PUT_RSP);
      }
      sst_throw(valueerror, "sst_message: invalid message type tostr %d", t);
  }

  const char*
  sst_message::tostr(mutation_t mut)
  {
      switch(mut)
      {
        casestr(RDMA_GET_REQ_TO_RSP);
        casestr(RDMA_PUT_REQ_TO_RSP);
        casestr(GNI_HEADER_TO_RDMA_GET_REQ);
        casestr(GNI_HEADER_TO_RDMA_PUT_REQ);
      }
      sst_throw(valueerror, "sst_message: invalid mutation type tostr %d", mut);
  }

} // end of namespace sstmac
