

/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_SIM_PARAMETERS_H_INCLUDED
#define SSTMAC_COMMON_SIM_PARAMETERS_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/units.h>
#include <sstmac/common/debug.h>
#include <vector>

#include <sstmac/common/basicstringtokenizer.h>
#include <boost/unordered_map.hpp>
#include <boost/foreach.hpp>

namespace sstmac{
  class sim_parameters : public ptr_type {
  protected:
    sim_parameters()
      {
      if(!dbg_.is_real()) {
          dbg_ = debug("<debug> parameters | params");
          logger::register_logger(&dbg_);
      }
    }

  public:
    typedef boost::intrusive_ptr<sim_parameters> ptr;

    virtual std::string
    get_param(std::string key) const = 0;
    
    
    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    virtual std::string
    get_optional_param(const std::string &key, const std::string &def) const {
      if (has_param(key)) return get_param(key);
      return def;
    }

    virtual void
    add_param(std::string key, std::string val) = 0;

    virtual void
    add_param_override(std::string key, std::string val) = 0;

    virtual void
    combine_into(sim_parameters::ptr sp) = 0;

    static std::pair<std::string, std::string>
     parse_param(std::string p)
     {
       std::string lhs = p.substr(0, p.find("="));
       std::string rhs = p.substr(p.find("=") + 1);

       if(lhs.find("**.") != std::string::npos)
           lhs = lhs.substr(3);

       lhs = sstmac::trim_str(lhs);
       rhs = sstmac::trim_str(rhs);

       return std::make_pair<std::string, std::string>(lhs, rhs);
     }

    virtual bool
    has_param(std::string key) const = 0;

    virtual int
    get_int_param(std::string key) const{
      std::string v = get_param(key);
      return atoi(v.c_str());
    }

    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    virtual int
    get_optional_int_param(const std::string &key, int def) const {
      if (has_param(key)) return get_int_param(key);
      return def;
    }

    /// Returns the value of the key as a boolean.
    virtual bool
    get_bool_param(const std::string &key) const {
      std::string v = get_param(key);
      if (v == "true" || v == "1") {
        return true;
      }
      else if (v != "false" && v != "0") {
        throw inputerror("bad bool value for " + key + ": " + v);
      }
      return false;
    }
    
    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    virtual bool
    get_optional_bool_param(const std::string &key, int def) const {
      if (has_param(key)) return get_bool_param(key);
      return def;
    }

    virtual long
    get_long_param(std::string key) const {
        std::string v = get_param(key);
        return atol(v.c_str());
    }
    
    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    virtual long
    get_optional_long_param(const std::string &key, long def) const {
      if (has_param(key)) return get_long_param(key);
      return def;
    }

    virtual double
    get_double_param(std::string key) const{
      std::string v = get_param(key);
      return atof(v.c_str());
    }
    
    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    virtual double
    get_optional_double_param(const std::string &key, double def) const {
      if (has_param(key)) return get_double_param(key);
      return def;
    }

    virtual timestamp
    get_time_param(std::string key) const {
      bool error = false;
      return get_timestamp(get_param(key).c_str(), error);

    }
    
    /// Return the value of the keyword if it exists. Otherwise return
    /// a default value.
    /// @param key gives the keyword
    /// @param def gives the default value (used if has_param(key) is false)
    /// @return the value if it exists, otherwise the default
    virtual timestamp
    get_optional_time_param(const std::string &key, timestamp def) const {
      if (has_param(key)) return get_time_param(key);
      return def;
    }

    void
    get_vector_param(const std::string& key,std::vector<int>& vals);

  protected:
     static sstmac::debug dbg_;

  };


  class empty_parameters : public sstmac::sim_parameters
    {
    protected:
    empty_parameters(){ }

        typedef boost::unordered_map<std::string, std::string> maptype;

    public:
      typedef boost::intrusive_ptr<empty_parameters> ptr;

      static ptr
      construct()
      {
        return ptr(new empty_parameters());
      }



      virtual std::string
      get_param(std::string key) const
      {
        SSTMAC_DEBUG << "file_parameters: getting param " << key << "\n";
        maptype::const_iterator it = params_.find(key);
        if (it == params_.end())
          {
            //std::cout << "parameter named: " << key << std::endl;
            sst_throw(sstmac::valueerror, "file_parameters: could not find parameter %s", key.c_str());
            //throw sstmac::valueerror("file_parameters: could not find parameter %s");
          }
        return it->second;
      }

      virtual bool
      has_param(std::string key) const
      {
        maptype::const_iterator it = params_.find(key);
        return (it != params_.end());
      }

      virtual void
      add_param(std::string key, std::string val)
      {
        SSTMAC_DEBUG << "setting key " << key << " to value " << val << "\n";

        if (params_.find(key) != params_.end())
          {
            throw sstmac::valueerror(
                "file_parameters::add_param - key already in params");
          }
        else
          {
            params_[key] = val;
          }
      }

      virtual void
      add_param_override(std::string key, std::string val)
      {
        SSTMAC_DEBUG << "setting key " << key << " to value " << val << "...";

        if (params_.find(key) != params_.end())
          {
            SSTMAC_DEBUG << "overriding \n";
          }
        else
          {
            SSTMAC_DEBUG << "adding new param \n";
          }

        params_[key] = val;

      }

      virtual void
      combine_into(sim_parameters::ptr sp)
      {
        BOOST_FOREACH(maptype::value_type it, params_)
                {
                  sp->add_param_override(it.first, it.second);
                }
      }

      virtual std::string
      to_string() const
      {
        return "empty_parameters";
      }

      virtual
      ~empty_parameters()
      {
      }

    protected:
      maptype params_;
    };

class sstmac_env {
    public:
        static sim_parameters::ptr params;
};

}

#endif
