#include <sstmac/hardware/memory/numa_memorymodel.h>
#include <sstmac/common/messages/sst_metadata.h>

namespace sstmac
{
  namespace hw
  {


    numa_memorymodel::~numa_memorymodel()
    {
    }

    void
    numa_memorymodel::handle(const sst_message::ptr &msg)
    {
      if(msg->get_type() == sst_message::FAILURE){
          failure_occurred(msg, ptr(this));
          return;
      }
   }

    int
    numa_memorymodel::get_socket(long threadnum) const
    {
        /** For now, just assume the NIC is hooked up
            to socket 0 */
        if (threadnum == sw::thread::nic_thread)
            return 0;
        /** Also assume the RDMA engine is hooked up
            to socket 0 */
        else if (threadnum == sw::thread::rdma_thread)
            return 0;

        int socket = sched_->get_socket(threadnum);
        if (socket == sw::thread::no_socket_affinity)
            return 0;
        return socket;
    }

    void
    numa_memorymodel::finish()
    {
        done_ = 0;
        memorymodel::finish();
    }

    void
    numa_memorymodel::do_access(const sst_message::ptr& msg)
    {
      use_me();

      mem_bytes_->record(msg->get_byte_length());

      int dst_socket = get_socket(msg->dest_thread(get_os()));
      int src_socket = get_socket(msg->source_thread(get_os()));

      hw::flow::ptr new_flow = flow::construct(msg, eventmodule::ptr(this), done_, flow_nsegments_, flow_nbytes_per_segment_, msg->get_event_time());
      if (src_socket != dst_socket){
          flow::link::ptr ht_link = socket_links_[src_socket][dst_socket];
          if (!ht_link){
              throw illformederror("numa memory model cannot yet handle anything but a fully connected set of sockets");
          }
          new_flow->add_link(ht_link);
      }
      flow::link::ptr bus_link = bus_links_[dst_socket];
      new_flow->add_link(bus_link);

      new_flow->start();
    }

    void
    numa_memorymodel::set_eventmanager(const eventmanager::ptr &m)
    {
        eventmodule::set_eventmanager(m);
    }

    sw::operatingsystem::ptr
    numa_memorymodel::get_os() const
    {
        return parent_node_->get_os();
    }

    void
    numa_memorymodel::read(const sst_message::ptr &msg)
    {
        SSTMAC_DEBUG << "NUMA memory model on node " << nodeid_
                     << " doing read of " << msg->get_byte_length()
                     << " bytes from socket " << get_socket(msg->source_thread(parent_node_->get_os()))
                     << " to socket " << get_socket(msg->dest_thread(get_os()))
                     << " on thread " << msg->dest_thread(get_os())
                     << "\n";
        do_access(msg);
    }

    void
    numa_memorymodel::write(const sst_message::ptr &msg)
    {
        SSTMAC_DEBUG << "NUMA memory model doing write of " << msg->get_byte_length() << "\n";
        do_access(msg);
    }

    void
    numa_memorymodel::rdma_get(const sst_message::ptr &msg)
    {
        SSTMAC_DEBUG << "NUMA memory model doing RDMA read of " << msg->get_byte_length() << "\n";
        do_access(msg);
    }

    void
    numa_memorymodel::rdma_put(const sst_message::ptr &msg)
    {
        SSTMAC_DEBUG << "NUMA memory model doing RDMA write of " << msg->get_byte_length() << "\n";
        do_access(msg);
    }

    void
    numa_memorymodel::nic_read(const sst_message::ptr& msg)
    {
        hw::flow::ptr new_flow = flow::construct(msg, ptr(this), done_, flow_nsegments_, flow_nbytes_per_segment_);

        int src_socket = get_socket(msg->source_thread(get_os()));
        flow::link::ptr mem_link = bus_links_[src_socket];
        new_flow->add_link(mem_link);

        //assume for now that nic goes to die 0 and then to memory
        if (src_socket != 0){
            flow::link::ptr ht_link = socket_links_[src_socket][0];
            if (!ht_link){
                throw illformederror("numa memory model cannot yet handle anything but a fully connected set of sockets");
            }
            new_flow->add_link(ht_link);
        }

        new_flow->add_link(nic_link_);
        new_flow->start();
    }

    void
    numa_memorymodel::nic_write(const sst_message::ptr &msg)
    {
        int dst_socket = get_socket(msg->dest_thread(get_os()));
        hw::flow::ptr new_flow = flow::construct(msg, ptr(this), done_, flow_nsegments_, flow_nbytes_per_segment_);
        new_flow->add_link(nic_link_);
        //assume for now that nic does to die and then to memory
        if (dst_socket != 0){
            flow::link::ptr ht_link = socket_links_[0][dst_socket];
            if (!ht_link){
                throw illformederror("numa memory model cannot yet handle anything but a fully connected set of sockets");
            }
            new_flow->add_link(ht_link);
        }
        flow::link::ptr mem_link = bus_links_[dst_socket];
        new_flow->add_link(mem_link);
        new_flow->start();
    }

    void
    numa_memorymodel::add_hypertransport_link(int src_socket, int dst_socket, double bw)
    {
        //assume symmetric bidirectional for now
        flow::link::ptr out_link = flow::link::construct(bw);
        socket_links_[src_socket][dst_socket] = out_link;
        flow::link::ptr in_link = flow::link::construct(bw);
        socket_links_[dst_socket][src_socket] = in_link;
    }

    numa_memorymodel::numa_memorymodel(
        const node::ptr& parent_node,
        const eventhandler::ptr &whendone,
        const timestamp &activate_latency,
        double bus_bw,
        double mem_bw,
        const timestamp& nic_link_lat,
        double nic_link_bw,
        int nsocket,
        long nodeid,
        int flow_nsegments,
        long flow_nbytes_per_segment,
        const hw::computescheduler::ptr& sched)
        :  memorymodel(parent_node, nodeid),
           done_(whendone),
           socket_links_(nsocket),
          bus_links_(nsocket),
           nsocket_(nsocket),
          sched_(sched),
          nodeid_(nodeid),
          flow_nsegments_(flow_nsegments),
          flow_nbytes_per_segment_(flow_nbytes_per_segment)
      {
        for (int i=0; i < nsocket; ++i){
            bus_links_[i] = flow::link::construct(bus_bw, mem_bw, activate_latency);
        }
        nic_link_ = flow::link::construct(nic_link_bw, nic_link_bw, nic_link_lat);
      }

  }
} /* namespace sstmac */
