/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/switches/circuit/circuitswitch.h>
#include <sstmac/hardware/network/switches/circuit/circuitmessage.h>

namespace sstmac
{
  namespace hw
  {

    stat_sum::ptr circuitswitch::timeout_stat_;

    circuitswitch::circuitswitch()
    {
      if (!timeout_stat_)
        {
          timeout_stat_ = stat_sum::construct("<stats> timeout",
              "circuit_switch_timeouts.csv");
          register_stat(timeout_stat_);
        }
    }

    networkswitch::ptr
    circuitswitch::clone(router::ptr route, nodeaddress::ptr addr)
    {
      ptr ret = construct();
      ret->init(addr, route);
      ret->circrouter_ = boost::dynamic_pointer_cast<circuit_router>(
          ret->router_);
      if (!ret->circrouter_)
        {
          throw ssterror(
              "circuitswitch: router didn't cast to a circuit_router");
        }
      ret->circrouter_->set_circ_switching();
      return ret;
    }

    injector::ptr
    circuitswitch::get_new_injector(sim_parameters::ptr params)
    {

      ptr sel(this);
      return circuitpacketinjector::construct(params, sel);

    }

    void
    circuitswitch::circuitpacketinjector::handle(const sst_message::ptr &msg)
    {
      timestamp delay(0);
      int vc;
      if (circpar_->numVC_ > 1)
        {
          circuitmessage::ptr circ =
              boost::dynamic_pointer_cast<circuitmessage>(msg);
          if (circ->get_circ_type() == circuitmessage::TEARDOWN
              || circ->get_circ_type() == circuitmessage::PATH_ACK)
            {
              vc = circpar_->numVC_ - 1;
            }
          else
            {
              vc = rand() % (circpar_->numVC_ - 1);
            }
        }
      else
        {
          vc = 1;
        }
      int inport;

      if (circpar_->nodebuffers_.find(msg->fromaddr_)
          == circpar_->nodebuffers_.end())
        {
          SSTMAC_DEBUG << "circuitswitch(" << parent_->get_addr()->to_string()
                << ")::packetinjector - adding node inport "
                << (circpar_->nodebufcnt_ - 1) << " for address "
                << msg->fromaddr_->to_string() << "\n";
          std::map<nodeaddress::ptr, int>::iterator it, end =
              circpar_->nodebuffers_.end();
          for (it = circpar_->nodebuffers_.begin(); it != end; it++)
            {
              SSTMAC_DEBUG << "in map: " << it->first->to_string() << "\n";
              SSTMAC_DEBUG << "equals: " << it->first->equals(msg->fromaddr_)
                    << "\n";
              SSTMAC_DEBUG << "==: " << (it->first == msg->fromaddr_) << "\n";
            }

          circpar_->nodebuffers_[msg->fromaddr_] = --circpar_->nodebufcnt_;
        }
      inport = circpar_->nodebuffers_[msg->fromaddr_];

      if (MTU_ > 0)
        {
          SSTMAC_DEBUG << "packetinjector: injecting packets: "
                << msg->to_string() << "\n";
          long bytes_sent = 0;
          long bytes_left = msg->get_byte_length();
          while (bytes_left > 0)
            {
              bytes_sent = std::min(MTU_, bytes_left);
              cycleaccuratemessage::ptr pack = cycleaccuratemessage::construct(
                  msg, bytes_sent, (bytes_sent >= bytes_left), vc);
              pack->inport_ = inport;
              parent_->schedule(parent_->now() + delay, parent_, pack);
              bytes_left -= bytes_sent;
              delay = delay + timestamp(bytes_sent / n2r_);
            }
        }
      else
        {
          delay = timestamp(msg->get_byte_length() / n2r_);
          SSTMAC_DEBUG << "packetinjector: injecting whole message: "
                << msg->to_string() << " with delay " << delay
                << " msg size is (" << msg->get_byte_length()
                << ") and bw is (" << n2r_ << ") \n";
          cycleaccuratemessage::ptr pack = cycleaccuratemessage::construct(msg,
              msg->get_byte_length(), true, vc);
          pack->inport_ = inport;
          parent_->schedule(parent_->now() + delay, parent_, pack);
        }

      if (msg->get_needs_ack())
        {
          sst_message::ptr sendack = msg->clone(sst_message::SEND_ACK);
          parent_->schedule(parent_->now() + delay,
              sendhandle_[msg->fromaddr_], sendack);
        }

    }

    bool
    circuitswitch::check_circuit(cycleaccuratemessage::ptr msg, int in,
        int out, bool &kill)
    {
      circuitmessage::ptr circ = boost::dynamic_pointer_cast<circuitmessage>(
          msg->orig());

      int realout = out;
      if (out < 0)
        realout = -1 * (circ->toaddr_->unique_id() + 1);

      if (circ->get_circ_type() == circuitmessage::SETUP)
        {
          if (circ->ttl_ < now())
            {
              kill = true;
              timeout_stat_->record(1, my_addr_->to_string());
              return false;
            }
          if (circuitsin_.find(in) == circuitsin_.end())
            {
              if (circuitsout_.find(realout) == circuitsout_.end())
                {

                  return true;
                }
              else
                {
                  SSTMAC_DEBUG << "circuitswitch[" << my_addr_->to_string()
                        << "]: circuit already setup on outport from "
                        << circuitsout_[realout] << " to " << realout << "\n";

                  return false;
                }
            }
          else
            {
              SSTMAC_DEBUG << "circuitswitch[" << my_addr_->to_string()
                    << "]: circuit already setup on inport from " << in
                    << " to " << circuitsin_[in] << "\n";

              return false;
            }
        }
      else if (circ->get_circ_type() == circuitmessage::TEARDOWN)
        {
          if (circuitsin_.find(in) == circuitsin_.end())
            {
              //throw ssterror(
              //    "circuitswitch::check_circuit -- trying to teardown a circuit that's not there in circuitsin");
              kill = true;
              return false;
            }

          if (circuitsout_.find(realout) == circuitsout_.end())
            {
              // throw ssterror(
              //    "circuitswitch::check_circuit -- trying to teardown a circuit that's not there in circuitsout");
              kill = true;
              return false;
            }

          if (circuitsin_[in] != realout || circuitsout_[realout] != in)
            {
              std::cout << "in = " << in << "\n";
              std::cout << "out = " << realout << "\n";
              std::cout << "circuitsin[" << in << "] = " << circuitsin_[in]
                  << "\n";
              std::cout << "circuitsout[" << realout << "] = "
                  << circuitsout_[realout] << "\n";
              throw ssterror(
                  "circuitswitch::check_circuit -- TEARDOWN -- circuit port information doesn't line up");

            }

          return true;

        }
      else if (circ->get_circ_type() == circuitmessage::BLOCKED)
        {
          if (circuitsin_.find(realout) == circuitsin_.end())
            {
              throw ssterror(
                  "circuitswitch::check_circuit -- trying to RELEASE a circuit that's not there in circuitsin");
            }

          if (circuitsout_.find(in) == circuitsout_.end())
            {
              throw ssterror(
                  "circuitswitch::check_circuit -- trying to RELEASE a circuit that's not there in circuitsout");
            }

          if (circuitsin_[realout] != in || circuitsout_[in] != realout)
            {
              std::cout << "in = " << in << "\n";
              std::cout << "out = " << realout << "\n";
              std::cout << "circuitsin[" << realout << "] = "
                  << circuitsin_[realout] << "\n";
              std::cout << "circuitsout[" << in << "] = " << circuitsout_[in]
                  << "\n";
              throw ssterror(
                  "circuitswitch::check_circuit -- BLOCKED -- circuit port information doesn't line up");
            }
          return true;
        }
      else if (circ->get_circ_type() == circuitmessage::TURNING)
        {
          return true;
        }
      else if (circ->get_circ_type() == circuitmessage::PATH_ACK)
        {
          return true;
        }
      else
        {
          throw ssterror(
              "circuitswitch::check_circuits() - unknown message type");
        }
    }

    void
    circuitswitch::finish()
    {
      cycleaccurateswitch::finish();
      if (circuitsin_.size() > 0)
        {
          SSTMAC_DEBUG << "circuitswitch[" << my_addr_->to_string()
                << "]: circuits still set up on finish \n";

          maptype::iterator it, end = circuitsin_.end();
          for (it = circuitsin_.begin(); it != end; it++)
            {
              SSTMAC_DEBUG << "circuit in: " << it->first << "->" << it->second
                    << "\n";
            }

          end = circuitsout_.end();
          for (it = circuitsout_.begin(); it != end; it++)
            {
              SSTMAC_DEBUG << "circuit out: " << it->first << "->"
                    << it->second << "\n";
            }
        }
    }

    void
    circuitswitch::do_route(cycleaccuratemessage::ptr msg, int inport,
        int &outport, bool &kill)
    {

      circuitmessage::ptr circ = boost::dynamic_pointer_cast<circuitmessage>(
          msg->orig());

      if (!circ)
        {
          throw valueerror("circuit received a non-circuit message");
        }

      if (circ->get_circ_type() == circuitmessage::SETUP
          || circ->get_circ_type() == circuitmessage::PATH_ACK)

        {
          SSTMAC_DEBUG << "routing circ type: " << circ->get_circ_type()
                << "\n";
          nodeaddress::ptr addr = circ->toaddr_;
          nodeaddress::ptr from = circ->fromaddr_;
          int dim;
          int dir;
          router_->route(routing_info::construct(addr), dim, dir);

          outport = router_->convert_to_outport(dim, dir);
        }
      else if (circ->get_circ_type() == circuitmessage::TEARDOWN)
        {

          if (circuitsin_.find(inport) == circuitsin_.end())
            {
              SSTMAC_DEBUG << "routing teardown by killing it \n";
              // std::cout << "inport: " << inport << "\n";
              // throw ssterror(
              //     "circuitswitch::do_route -- trying to find a circuit that's not there in circuitsin");
              kill = true;
              outport = -1;
              return;
            }

          SSTMAC_DEBUG
                << "routing teardown by grabbing the existing circuit from inport "
                << inport << "\n";
          outport = circuitsin_[inport];
          SSTMAC_DEBUG << "raw: " << outport << "\n";
          if (outport < 0)
            outport = -1;

          SSTMAC_DEBUG << "returning: " << outport << "\n";

        }
      else if (circ->get_circ_type() == circuitmessage::BLOCKED)
        {
          SSTMAC_DEBUG
                << "routing blocked message by grabbing the reverse circuit \n";
          if (circuitsout_.find(inport) == circuitsout_.end())
            {
              std::cout << "inport: " << inport << "\n";
              throw ssterror(
                  "circuitswitch::do_route -- trying to find a circuit that's not there in circuitsout");
            }

          outport = circuitsout_[inport];
          if (outport < 0)
            outport = -1;
        }
      else if (circ->get_circ_type() == circuitmessage::TURNING)
        {
          SSTMAC_DEBUG << "routing turning by returning the inport \n";
          outport = inport;

        }
      else
        {
          throw ssterror("circuitswitch::do_route() - unknown message type");
        }

    }

    void
    circuitswitch::modify_circuits(cycleaccuratemessage::ptr msg, int in,
        int out)
    {
      circuitmessage::ptr circ = boost::dynamic_pointer_cast<circuitmessage>(
          msg->orig());

      int realout = out;
      if (out < 0)
        realout = -1 * (circ->toaddr_->unique_id() + 1);

      if (circ->get_circ_type() == circuitmessage::SETUP)
        {

          SSTMAC_DEBUG << "circuitswitch[" << my_addr_->to_string()
                << "]: setting up circuit from " << in << " to " << realout
                << "\n";
          circuitsin_[in] = realout;
          circuitsout_[realout] = in;

          circrouter_->path_is_good(circ->toaddr_, in, out);
        }
      else if (circ->get_circ_type() == circuitmessage::TEARDOWN)
        {

          SSTMAC_DEBUG << "circuitswitch[" << my_addr_->to_string()
                << "]: tearing down circuit from " << in << " to "
                << circuitsin_[in] << " from a TEARDOWN message \n";
          circuitsin_.erase(in);
          circuitsout_.erase(realout);

          circrouter_->path_teardown(in, out);

        }
      else if (circ->get_circ_type() == circuitmessage::BLOCKED)
        {

          SSTMAC_DEBUG << "circuitswitch[" << my_addr_->to_string()
                << "]: tearing down circuit from " << realout << " to "
                << circuitsin_[realout] << " from a BLOCKED message \n";
          circuitsin_.erase(realout);
          circuitsout_.erase(in);

          circrouter_->path_teardown(out, in);
        }
      else if (circ->get_circ_type() == circuitmessage::TURNING)
        {
          circ->set_circ_type(circuitmessage::BLOCKED);
          nodeaddress::ptr addr = circ->toaddr_;
          circ->toaddr_ = circ->fromaddr_;
          circ->fromaddr_ = addr;
        }
      else if (circ->get_circ_type() == circuitmessage::PATH_ACK)
        {

        }
      else
        {
          throw ssterror(
              "circuitswitch::modify_circuits() - unknown message type");
        }
    }

    int
    circuitswitch::check_inport(int inport, bool &clock)
    {
      int ret = 0;

      /* SSTMAC_DEBUG << "switch[" << my_addr_->to_string()
       << "]: examining inport " << inport << ", num_in_buff = "
       << inbuffcnt_[inport] << ", time = " << now() << ", inport_free = "
       << inports_free_[inport].sec() << "\n";*/

      if (inbuffcnt_[inport] > 0)
        {
          if (inports_free_[inport] <= now())
            {

              int beginVC = (arb_vc_[inport] + 1) % numVC_;
              bool firstVC = true;
              for (int vc = beginVC; firstVC || vc != beginVC; vc = (vc + 1)
                  % numVC_)
                {

                  firstVC = false;
                  int index = inport * numVC_ + vc;

                  SSTMAC_DEBUG << "switch[" << my_addr_->to_string()
                        << "]: examining vc " << vc << ", with index " << index
                        << ", queue size = " << inbuffs_[index].size() << "\n";

                  if (!inbuffs_[index].empty())
                    {
                      cycleaccuratemessage::ptr msg = inbuffs_[index].front();
                      // SSTMAC_DEBUG << "got message: " << msg->to_string()
                      //       << "\n";

                      circuitmessage::ptr circ = boost::dynamic_pointer_cast<
                          circuitmessage>(msg->orig());

                      int outport;
                      bool kill = false;
                      do_route(msg, inport, outport, kill);

                      if (kill)
                        {
                          SSTMAC_DEBUG << "killing message "
                                << msg->to_string() << "\n";
                          inbuffs_[inport * numVC_ + vc].pop();
                          inbuffcnt_[inport]--;
                          arb_vc_[inport] = vc;
                          clock = true;

                          if (inport >= 0)
                            {
                              //send credit ack back
                              cycleaccuratemessage::ptr ack =
                                  cycleaccuratemessage::construct(
                                      msg->get_byte_length(), vc);

                              ack->inport_ = msg->prevport_;

                              SSTMAC_DEBUG << "ca-switch("
                                    << my_addr_->to_string()
                                    << "): sending ack back to "
                                    << msg->prevswitch_->to_string() << "\n";

                              schedule(now() + link_lat_, msg->prevswitch_, ack);
                            }
                          break;
                        }

                      if (msg->vc_ != vc)
                        {
                          throw ssterror(
                              "cycleaccurate switch: vcs don't match for some reason");
                        }

                      /* SSTMAC_DEBUG << "packet going to outport" << outport
                       << ", outports_free = "
                       << outports_free_[outport].sec();*/

                      /* if (outport >= 0)
                       {
                       SSTMAC_DEBUG << ", outbuff occupancy = "
                       << outbuffs_[outport]->get_occupancy()
                       << ", credits = " << credits_[outport * numVC_
                       + vc] << "\n";
                       }
                       else SSTMAC_DEBUG << "\n";*/

                      if (outports_free_[outport] <= now() && (outport < 0
                          || (outBuffSize_
                              - outbuffs_[outport]->get_occupancy()
                              >= msg->get_byte_length() && credits_[outport
                              * numVC_ + vc] >= mtu)))
                        {

                          bool checked = check_circuit(msg, inport, outport,
                              kill);

                          if (checked)
                            {
                              SSTMAC_DEBUG << "switch["
                                    << my_addr_->to_string() << "]: message "
                                    << msg->to_string()
                                    << " arbitration good from " << inport
                                    << " to " << outport << " \n";

                              modify_circuits(msg, inport, outport);

                              inbuffs_[inport * numVC_ + vc].pop();
                              inbuffcnt_[inport]--;
                              //now we're ok to send through the crossbar
                              //  sendDelayed(msg, arbLatency_, outports[outport]);
                              timestamp outbuffarrival =
                                  now() + switchLatency_
                                      + timestamp(
                                          msg->get_byte_length()
                                              / switchBandwidth_);
                              int prevport = msg->prevport_;
                              networkswitch::ptr prevswitch = msg->prevswitch_;
                              msg->prevswitch_ = ptr(this);
                              msg->prevport_ = outport;

                              // circ->stack_.push(self_);
                              schedule(outbuffarrival, outbuffs_[outport], msg);

                              if (using_orion_)
                                {
                                  //power for leaving inport
                                  power_stat_->record(
                                      now(),
                                      power_buffer_in_->stat_energy(
                                          msg->get_byte_length() / line_width_,
                                          0, 0));

                                  //power for entering crossbar
                                  double initial = power_xbar_->report();
                                  power_xbar_->record(1, inport, outport,
                                      msg->get_byte_length());
                                  power_stat_->record(now(),
                                      power_xbar_->report() - initial);
                                }

                              inports_free_[inport] = outbuffarrival;
                              outports_free_[outport] = outbuffarrival;
                              arb_vc_[inport] = vc;
                              ret++;
                              clock = true;

                              if (outport >= 0)
                                {
                                  int inbuffindex = outport * numVC_ + vc;
                                  credits_[inbuffindex]
                                      -= msg->get_byte_length();
                                }

                              if (inport >= 0)
                                {
                                  //send credit ack back
                                  cycleaccuratemessage::ptr ack =
                                      cycleaccuratemessage::construct(
                                          msg->get_byte_length(), vc);

                                  ack->inport_ = prevport;

                                  SSTMAC_DEBUG << "ca-switch("
                                        << my_addr_->to_string()
                                        << "): sending ack back to "
                                        << prevswitch->to_string() << "\n";

                                  schedule(now() + link_lat_, prevswitch, ack);
                                }

                              break;

                            }
                          else if (kill)
                            {
                              SSTMAC_DEBUG << "killing message "
                                    << msg->to_string() << "\n";
                              inbuffs_[inport * numVC_ + vc].pop();
                              inbuffcnt_[inport]--;
                              arb_vc_[inport] = vc;
                              clock = true;

                              if (inport >= 0)
                                {
                                  //send credit ack back
                                  cycleaccuratemessage::ptr ack =
                                      cycleaccuratemessage::construct(
                                          msg->get_byte_length(), vc);

                                  ack->inport_ = msg->prevport_;

                                  SSTMAC_DEBUG << "ca-switch("
                                        << my_addr_->to_string()
                                        << "): sending ack back to "
                                        << msg->prevswitch_->to_string()
                                        << "\n";

                                  schedule(now() + link_lat_, msg->prevswitch_,
                                      ack);
                                }
                              break;
                            }
                          else if (blocked_protocol_)
                            {

                              if (circ->get_circ_type()
                                  == circuitmessage::SETUP)
                                { //uh oh, gotta turn it around
                                  SSTMAC_DEBUG << "circuitswitch("
                                        << my_addr_->to_string()
                                        << "): turning this boat around \n";
                                  circ->set_circ_type(circuitmessage::TURNING);
                                  clock = true;
                                }
                            }

                        }
                    }
                }
            }
        }

      return ret;
    }

  }
}
