/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/switches/packet/packetswitch.h>
#include <sstmac/hardware/network/switches/packet/packet_routing_info.h>

namespace sstmac
{
  namespace hw
  {

    stat_time_window_avg::ptr packetswitch::stat_Q_wait_ = NULL;
    stat_time_window_sum::ptr packetswitch::stat_bytes_ = NULL;
    stat_time_window_sum_duration::ptr packetswitch::stat_pwr_ = NULL;

    using namespace std;

    networkswitch::ptr
    packetswitch::clone(router::ptr route, nodeaddress::ptr addr)
    {
      ptr ret = construct();
      ret->init(addr, route);
      return ret;
    }

    injector::ptr
    packetswitch::get_new_injector(sim_parameters::ptr params)
    {
      ptr self(this);
      return packetinjector::construct(params, self);

    }

    ejector::ptr
    packetswitch::get_new_ejector(sim_parameters::ptr params)
    {
      ptr self(this);
      return packetejector::construct(params, self, my_addr_);
    }

    void
    packetswitch::initialize(sim_parameters::ptr params, int nps)
    {
      if (!stat_Q_wait_)
        stat_Q_wait_ = stat_time_window_avg::construct("<stats> network",
            "stats_packet_Qwait.csv");

      if (!stat_bytes_)
        stat_bytes_ = stat_time_window_sum::construct("<stats> network",
            "stats_packet_bytes.csv");

      if (!stat_pwr_)
        stat_pwr_ = stat_time_window_sum_duration::construct("<stats> network",
            "stats_power_network.csv");

      lat_r2r_ = params->get_time_param("packetswitch_latency_r2r");

      for (int i = 0; i < nps; i++)
        {
          portbw_[EJECT_PORT][i] = get_bandwidth(
              params->get_param("packetswitch_bandwidth_n2r").c_str());
        }

      redundRNG_ = RNG::SimpleCombo::construct(time(NULL));
      srand(time(NULL));


      maxtime_ = timestamp(0);
    }

    void
    packetswitch::handle(const sst_message::ptr &msg)
    {

      packetmessage::ptr pack = boost::dynamic_pointer_cast<packetmessage>(msg);

      if (!pack)
        {
          throw ssterror("packetswitch::handle() - can't cast to packetmessage");
        }

      bool isDimChange = false;
      int dim;
      int dir;

      pack->rinfo()->set_congestion(finishtimes_, maxtime_, now());
      router_->route(pack->rinfo(), dim, dir);

      if (redund_.find(dim) == redund_.end())
        {
          redund_[dim] = std::map<int, int>();
        }

      if (redund_[dim].find(dir) == redund_[dim].end())
        {
          redund_[dim][dir] = 1;
        }

      if(redund_[dim][dir] < 1){
          redund_[dim][dir] = 1;
      }
      SSTMAC_DEBUG << "dim: " << dim << ", dir: " << dir << ", redund: " << redund_[dim][dir] << ", portbw: " << portbw_[dim][dir] << "\n";

      if (finishtimes_.find(dim) == finishtimes_.end())
        {
          std::map<int, std::map<int, eventhandler::ptr> >::iterator it, end =
              outports_.end();
          for (it = outports_.begin(); it != end; it++)
            {
              std::map<int, eventhandler::ptr>::iterator it2, end2 =
                  it->second.end();
              for (it2 = it->second.begin(); it2 != end2; it2++)
                {
                  for (int i = 0; i < redund_[it->first][it2->first]; i++)
                    {
                      finishtimes_[it->first][it2->first][i] = timestamp(0);
                    }
                }
            }

        }

      int r = rand() % redund_[dim][dir];
      timestamp txFinishTime = finishtimes_[dim][dir][r];

      if (txFinishTime < now())
        txFinishTime = now();
      timestamp delay = lat_r2r_ + timestamp(
          pack->get_byte_length() / portbw_[dim][dir]);

      SSTMAC_DEBUG << "message going to dim " << dim << " and dir " << dir
            << " which is switch " << outports_[dim][dir]->to_string()
            << " with delay " << delay << "\n";
      schedule(txFinishTime + delay, outports_[dim][dir], msg);

      finishtimes_[dim][dir][r] = txFinishTime + delay;

      if(finishtimes_[dim][dir][r] - now() > maxtime_){
          maxtime_ = finishtimes_[dim][dir][r] - now();
      }

      double wait = (txFinishTime.sec() - now().sec());
      stat_Q_wait_->record(txFinishTime, wait);

      stat_bytes_->record(txFinishTime, pack->get_byte_length());

    }

    void
    packetswitch::packetinjector::handle(const sst_message::ptr &msg)
    {
      timestamp delay(0);
      if (MTU_ > 0)
        {
          SSTMAC_DEBUG << "packetinjector: injecting packets: "
                << msg->to_string() << "\n";
          long bytes_sent = 0;
          long bytes_left = msg->get_byte_length();
          while (bytes_left > 0)
            {
              bytes_sent = std::min(MTU_, bytes_left);
              packetmessage::ptr pack = packetmessage::construct(msg,
                  bytes_sent, (bytes_sent >= bytes_left));
              parent_->schedule(parent_->now() + delay, parent_, pack);
              bytes_left -= bytes_sent;
              delay = delay + timestamp(bytes_sent / n2r_);
            }
        }
      else
        {
          delay = timestamp(msg->get_byte_length() / n2r_);
          SSTMAC_DEBUG << "packetinjector: injecting whole message: "
                << msg->to_string() << " with delay " << delay
                << " msg size is (" << msg->get_byte_length()
                << ") and bw is (" << n2r_ << ") \n";

          parent_->schedule(parent_->now() + delay, parent_,
              packetmessage::construct(msg, msg->get_byte_length(), true));
        }

      if (msg->get_needs_ack())
        {
          sst_message::ptr sendack = msg->clone(sst_message::SEND_ACK);
          parent_->schedule(parent_->now() + delay,
              sendhandle_[msg->fromaddr_], sendack);
        }

    }

    void
    packetswitch::packetejector::handle(const sst_message::ptr &msg)
    {
      packetmessage::ptr pack = boost::dynamic_pointer_cast<packetmessage>(msg);

      if (!pack)
        {
          throw ssterror(
              "packetswitch::packetejector::handle() - can't cast to packetmessage");
        }

      if (pack->is_tail())
        {
          if (recvhandles_.find(pack->orig()->toaddr_) == recvhandles_.end())
            {
              std::cout << "my addr: " << addr_->to_string() << "\n";
              std::cout << "message headed to: "
                  << pack->orig()->toaddr_->to_string() << "\n";
              std::cout << "message came from: "
                  << pack->orig()->fromaddr_->to_string() << "\n";
              throw ssterror(
                  "packetswitch::packetejector::handle() -- can't find recv handler, that's a problem");
            }
          SSTMAC_DEBUG << "packetejector: ejecting message to "
                << pack->orig()->toaddr_->to_string() << "\n";
          parent_->schedule(parent_->now(),
              recvhandles_[pack->orig()->toaddr_], pack->orig());
        }
    }

  }
}
