/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/switches/routing/closrouter.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
namespace sstmac
{
  namespace hw
  {

    std::map<int, int> closrouter::middlealloc_;
    std::map<int, int> closrouter::inchannelsalloc_;
    std::map<int, int> closrouter::outchannelsalloc_;

    void
    closrouter::route(routing_info::ptr to, int &dim, int &dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(to->to());

      SSTMAC_DEBUG << "closrouter(" << me_->to_string() << "): heading to "
            << to->to_string() << " at net addr " << netaddr->to_string()
            << "\n";

      if (me_->unique_id() >= m_ + k_) //i'm an output stage
        {
          dim = networkswitch::EJECT_PORT;
          dir = to->to()->unique_id() % conc_;
        }
      else if (me_->unique_id() >= m_)
        { // i'm a mid-stage switch
          dim = 0;
          dir = netaddr->unique_id();
        }
      else
        { //I'm an input stage switch

          dim = 0;
          if (!is_circuit_switching_)
            {
              dir = rand() % k_;
            }
          else
            {
              int start = rand() % k_;
              bool first = true;
              for (int k = start; first || k != start; k = (k + 1) % k_)
                {
                  first = false;
                  int index = k * m_ + me_->unique_id();
                  if (inchannelsalloc_[index] < channels_)
                    { // the input port is free for this switch
                      int outdex = k * m_ + netaddr->unique_id();
                      bool free = (outchannelsalloc_[outdex] < channels_);
                     /* for (int m = 0; m < m_; m++)
                        { //check the outports from this switch
                          int ind2 = k * m_ + m;
                          if (middlealloc_.find(ind2) != middlealloc_.end())
                            {
                              if (middlealloc_[ind2] == netaddr->unique_id())
                                {
                                  free = false;
                                }
                            }
                        }*/

                      if (free)
                        {
                          SSTMAC_DEBUG << "closrouter: returning dir " << k << "\n";
                          dir = k;
                          return;
                        }
                    }
                }

              SSTMAC_DEBUG << "closrouter: didn't find a free port \n";
              dir = 0;
            }

        }
      return;

    }

    bool
    closrouter::is_fixed_routing() const
    {
      return me_->unique_id() < m_;
    }

    int
    closrouter::convert_to_outport(int dim, int dir)
    {
      if (dim == networkswitch::EJECT_PORT)
        {
          return -(1+dir);
        }
      return dir;
    }

    void
    closrouter::convert_to_dimdir(int outport, int &dim, int&dir)
    {
      if (outport < 0)
        {
          dim = networkswitch::EJECT_PORT;
          dir = (-outport) - 1;
        }
      else
        {
          dim = 0;
          dir = outport;
        }
    }

    int
    closrouter::get_inport(nodeaddress::ptr from, int dim, int dir)
    {

      if (me_->unique_id() >= m_ + k_)
        {
          return from->unique_id() - m_;
        }
      else if (me_->unique_id() >= m_)
        {
          return from->unique_id();
        }
      else
        {
          return from->unique_id();
        }
    }

    void
    closrouter::path_is_good(nodeaddress::ptr goingto, int fromport, int toport)
    {
      if (me_->unique_id() < m_) // i'm an input stage
        {
          int outport = convert_to_network_addr(goingto)->unique_id();
          int index = toport * m_ + me_->unique_id();
          int outdex = toport * m_ + outport;
          SSTMAC_DEBUG << "closrouter[" << me_->unique_id()
                << "]: path is good at index " << index << " to outport "
                << outport << "\n";


          //middlealloc_[index] = outport;
          inchannelsalloc_[index]++;
          outchannelsalloc_[outdex]++;

        }
    }

    void
    closrouter::path_teardown(int fromport, int toport)
    {
      if (me_->unique_id() < m_ + k_ && me_->unique_id() >= m_) // i'm a middle-stage
        {
          int index = (me_->unique_id() - m_) * m_ + fromport;
          int outdex = (me_->unique_id() - m_) * m_ + toport;
          SSTMAC_DEBUG << "closrouter[" << me_->unique_id()
                << "]: path teardown from " << index << " to outport "
                << toport << "\n";

         // middlealloc_.erase(index);
          inchannelsalloc_[index]--;
          outchannelsalloc_[outdex]--;

        }
    }

  }
}
