/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/switches/routing/dragonflyrouter.h>
#include <math.h>
#include <sstmac/hardware/network/switches/networkswitch.h>

namespace sstmac
{
  namespace hw
  {

    void
    dragonflyrouter::route(routing_info::ptr to, int &dim, int &dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(to->to());
      int hisx, hisy, hisg;
      get_coords(netaddr->unique_id(), hisx, hisy, hisg);
      SSTMAC_DEBUG << "dragonflyrouter(" << me_->to_string()
            << "): heading to " << to->to()->to_string() << " at net addr "
            << netaddr->to_string() << "\n";

      if (hisg != myG_)
        { //route to the right group first

          int gspace = std::max(1, numG_ / 5);
          int myid = myX_ + myY_ * numX_;
          int gset = myid % numG_;
          int dist = 1000000;
          for (int g = 0; g < 5; g++)
            {

              int gid = (gset + gspace * g) % numG_;
              if (gid == hisg)
                {
                  dim = 2;
                  dir = gid;
                  return;
                }
              else
                {
                  if (std::abs(gid - hisg) < dist)
                    {
                      dist = std::abs(gid - hisg);
                    }
                }

            }

          //if we get here, we didn't find a group connection here
          dim = 0;
          dir = (myX_ + 1) % numX_;
        }
      else if (hisx != myX_)
        {
          dim = 0;
          dir = hisx;
        }
      else if (hisy != myY_)
        {
          dim = 1;
          dir = hisy;
        }
      else
        {
          dim = networkswitch::EJECT_PORT;
          dir = to->to()->unique_id() % conc_;
        }

    }

    int
    dragonflyrouter::convert_to_outport(int dim, int dir)
    {
      if (dim == networkswitch::EJECT_PORT)
        {
          return -(1+dir);
        }
      int ret = 0;
      if (dim == 0)
        {
          return dir;
        }
      else if (dim == 1)
        {
          return numX_ + dir;
        }
      else
        return numX_ + numY_ + dir;
    }

    void
    dragonflyrouter::convert_to_dimdir(int outport, int &dim, int&dir)
    {
      if (outport < 0)
        {
          dim = networkswitch::EJECT_PORT;
          dir = (-outport) - 1;
        }
      else
        {
          if (outport < numX_)
            {
              dim = 0;
              dir = outport;
            }
          else if (outport >= numX_ && outport < numX_ + numY_)
            {
              dim = 1;
              dir = outport - numX_;
            }
          else
            {
              dim = 2;
              dir = outport - (numX_ + numY_);
            }

        }
    }

    int
    dragonflyrouter::get_inport(nodeaddress::ptr from, int dim, int dir)
    {
      nodeaddress::ptr netaddr = convert_to_network_addr(from);
      int hisx, hisy, hisg;
      get_coords(netaddr->unique_id(), hisx, hisy, hisg);
      int thedir;
      if(dim == 0){
          thedir = hisx;
      }else if(dim ==1){
          thedir = hisy;
      }else{
          thedir = hisg;
      }
      return convert_to_outport(dim, thedir);

    }
  }

}

