/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/switches/switchinterconnect.h>

#include <sstmac/hardware/network/switches/packet/packetswitch.h>
#include <sstmac/hardware/network/switches/cycleaccurate/cycleaccurateswitch.h>
#include <sstmac/hardware/network/switches/circuit/circuitswitch.h>
#include <sstmac/hardware/network/switches/switchtopology.h>

#include <sstmac/common/factories/network_factory.h>


namespace sstmac
{
  namespace hw
  {
    SSTRegisterNetwork("switch", switchinterconnect);

    using namespace std;

    switchinterconnect::switchinterconnect(topology::ptr top,
        sim_parameters::ptr params, vis::vis_engine::ptr veng) :
      interconnect(top), vis_display("interconnect", veng)
    {
      switchtop_ = boost::dynamic_pointer_cast<switchtopology>(top);
      if (!switchtop_)
        {
          std::cout << "topolgy must support switch model, topology name: "
              << params->get_param("topology_name") << "\n";
          throw ssterror(
              "switchinterconnect::ctor -- topolgy must support switch model");
        }

      networkswitch::ptr switchtemp;
      std::string switchtype = params->get_param("network_switchtype");
      int nps = params->get_int_param("network_nodes_per_switch");
      SSTMAC_DEBUG << "switchinterconnect: using switch model: " << switchtype
            << "\n";
      if (switchtype == "packet")
        {
          switchtemp = packetswitch::construct();

        }
      else if (switchtype == "cycleaccurate")
        {
          switchtemp = cycleaccurateswitch::construct();
        }
      else if (switchtype == "circuit")
        {
          switchtemp = circuitswitch::construct();
        }

      for (long i = 0; i < top->n() * nps; i++)
        {
          nodeaddress::ptr nodeaddr = nodeaddress::convert_from_uniqueid(i);
          nodeaddress::ptr netaddr =
              nodeaddress::convert_from_uniqueid(i / nps);

          if (switches_.find(netaddr) == switches_.end())
            {
              networkswitch::ptr sw = switchtemp->clone(
                  switchtop_->create_router(netaddr, params), netaddr);
              switches_[netaddr] = sw;

            }

          injector_[nodeaddr] = switches_[netaddr]->get_injector(params,
              i % nps);

        }

      SSTMAC_DEBUG << "switchinterconnect: created " << switches_.size()
            << " gateways switches for " << top->n() << " nodes \n";

      switchtop_->connect_switches(switches_, params);

      //topology must set ejectors for the case of indirect networks
      switchtop_->set_ejectors(top->n(), nps, switches_, ejector_, params);

      //now they're connected up, initialize them.  Note that
      //  connect_switches could very well have added some switches
      switchtopology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
        {
          it->second->initialize(params, nps);
        }

    }

    /// Send the given message at the given time.
    /// Notify the sender node when the send is complete
    /// and notify the receiver node when receive needs to be performed.
    void
    switchinterconnect::send(const timestamp &start_time,
        const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "switchinterconnect: sending to "
            << msg->toaddr_->to_string() << " from "
            << msg->fromaddr_->to_string() << " using injector "
            << injector_[msg->fromaddr_]->to_string() << " and ejector "
            << ejector_[msg->toaddr_]->to_string() << "\n";

      // injector_[msg->fromaddr_]->set_sendhandler(endpoints_[msg->fromaddr_], msg->fromaddr_);
      // ejector_[msg->toaddr_]->set_recvhandler(endpoints_[msg->toaddr_], msg->toaddr_);
      schedule(start_time, injector_[msg->fromaddr_], msg);
    }

    void
    switchinterconnect::set_eventmanager(const eventmanager::ptr &m)
    {
      eventmodule::set_eventmanager(m);

      switchtopology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
        {
          it->second->set_eventmanager(m);
        }
    }

    void
    switchinterconnect::finish()
    {
      switchtopology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
        {
          it->second->finish();
        }
    }

    vis::vis_topology::vismaptype
    switchinterconnect::vis_get_nodes()
    {
      vis::vis_topology::vismaptype vismap;
      switchtopology::maptype::iterator it, end = switches_.end();
      for (it = switches_.begin(); it != end; it++)
        {
          vismap[it->first] = it->second;
        }
      return vismap;
    }

    vis::vis_topology::ptr
    switchinterconnect::vis_get_topology()
    {
      return switchtop_;
    }

  }
}
