/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// clos.h: Interface for CLOS networks.
//
// Based on clos.c by Jim Schutt
// Adapted by Curtis Janssen <cljanss@ca.sandia.gov>
#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_CLOS_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_CLOS_H_INCLUDED

#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/hardware/network/switches/switchtopology.h>

namespace sstmac
{
  namespace hw
  {

    /** The clos network generates a full 3-tier CLOS network with 2*n port
     switches.  In the CLOS network, not all vertices are nodes.

     <table border="1">

     <tr><td>Keyword<td>Type<td>Default<td>Description

     <tr><td><tt>n</tt><td>int<td>2<td>Half of the number of ports in each
     switching element.

     </table>

     */
    class clos : public switchtopology
    {
    private:
      int n_;
      int k_;
      int m_;

      bool nonswitch_;

      void
      route_hashed(const vertex &v1, const vertex &v2, walk &w) const;
      void
      init_clos(int n);

    protected:
      //clos(TiXmlHandle);
      clos(int n, int k, int m);

      clos(int n);

    public:
      typedef boost::intrusive_ptr<clos> ptr;

      static clos::ptr
      construct(sim_parameters::ptr params);

      static clos::ptr
      construct(int n = 2);

      virtual
      ~clos();

      virtual int
      radix() const
      {
        throw ssterror("clos doesn't have a fixed radix");
      }

      static size_t
      size(int n);
      void
      shortest_path(const vertex&v1, const vertex&v2, walk&w) const;
      void
      route(const vertex &v1, const vertex &v2, walk &w) const;
      virtual int64_t
      n() const;
      int64_t
      diameter() const;
      double
      a() const;
      int64_t
      b() const;
      bool
      vertex_symmetric() const;
      std::string
      name() const;

      virtual void
      connect_switches(maptype &switches, sim_parameters::ptr params,
          std::string linkbwname = "network_bandwidth_link");

      virtual void
      set_ejectors(long num, int nps, maptype switches, ejmaptype &ejectors,
          sim_parameters::ptr params);

      virtual router::ptr
      create_router(nodeaddress::ptr addr, sim_parameters::ptr params);

      virtual int64_t
      topology_size() const
      {

        return n();
      }

      virtual std::string
      to_string() const
      {
        return "clostopology";
      }
    };

  }
} //end of namespace sstmac

#endif
