/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
// crossbar.cc: Implementation of crossbar networks.
//
// Author: Curtis Janssen <cljanss@sandia.gov>
#include <sstream>
#include <sstmac/hardware/network/topology/crossbar.h>
#include <sstmac/hardware/network/switches/routing/crossbarrouter.h>
#include <sstmac/common/driver_util.h>
#include <sstmac/common/factories/topology_factory.h>


using std::set;
using std::vector;
using std::string;

namespace sstmac
{
  namespace hw
  {
    SSTRegisterTopology("crossbar | xbar", crossbar);

    crossbar::ptr
    crossbar::construct(sim_parameters::ptr params)
    {
      bool errorflag = false;
      std::vector<int> args = get_intvec(
          params->get_param("topology_geometry").c_str(), errorflag);
      int nps = params->get_optional_int_param("network_nodes_per_switch", 1);
      return crossbar::ptr(new crossbar(args[0], nps));
    }

    crossbar::crossbar(long nnode, int nps) :
      nps_(nps), size_(nnode)
    {
      nodes_.clear();
      for (int i = 0; i < nnode; i++)
        {
          std::vector<int> node(1);
          node[0] = i;
          nodes_.insert(node);
        }

      set<vertex> all_vertices(nodes_);
      std::vector<int> sw(1);
      sw[0] = -1;
      vertex swvertex(sw);
      all_vertices.insert(swvertex);

      set<edge> edges;
      for (set<vertex>::iterator i = nodes_.begin(); i != nodes_.end(); i++)
        {
          edge e1(*i, swvertex);
          edge e2(swvertex, *i);
          edges.insert(e1);
          edges.insert(e2);
        }

      g_.init(all_vertices, edges);
      sw_index_ = g_.vertex_to_index(sw);
      init();
    }

    double
    crossbar::a() const
    {
      return 2.0;
    }

    int64_t
    crossbar::b() const
    {
      return n() / 2;
    }

    void
    crossbar::shortest_path(const vertex &v1, const vertex &v2, walk &w) const
    {
      vertex_index v1i = g().vertex_to_index(v1);
      vertex_index v2i = g().vertex_to_index(v2);
      if (v1i == v2i)
        {
          w.evec().clear();
          return;
        }
      w.evec().resize(2);
      w.evec()[0].init(v1i, sw_index_);
      w.evec()[1].init(sw_index_, v2i);
    }

    int64_t
    crossbar::diameter() const
    {
      return 2;
    }

    int64_t
    crossbar::topology_size() const
    {
      return size_;
    }

    bool
    crossbar::vertex_symmetric() const
    {
      return true;
    }

    string
    crossbar::name() const
    {
      std::ostringstream ostr;
      ostr << "crossbar(" << n() << ")";
      return ostr.str();
    }

    bool
    crossbar::is_switch(vertex_index vi) const
    {
      return vi == sw_index_;
    }

    void
    crossbar::connect_switches(switchtopology::maptype &switches,
        sim_parameters::ptr params, std::string linkbwname)
    {
      double linkbw = get_bandwidth(params->get_param(linkbwname).c_str());
      for (long i = 0; i < switches.size(); i++)
        {
          nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);

          for (long j = 0; j < switches.size(); j++)
            {
              nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(j);
              if (i != j)
                {
                  switches[me]->connect(0, j, switches[them], linkbw);
                }

            }

        }

    }

    router::ptr
    crossbar::create_router(nodeaddress::ptr addr, sim_parameters::ptr params)
    {

      router::ptr ret = crossbarrouter::construct(addr, nps_);

      return ret;
    }

  }
} //end of namespace sstmac
