/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// hdtorus.cc: Implementation of a high dimensional torus network.
//
// Author: Ali Pinar <apinar@sandia.gov> 
// 

#include <sstream>
#include <sstmac/hardware/network/topology/hdtorus.h>
#include <sstmac/hardware/network/switches/routing/torusrouter.h>
// #include <sstmac/network/ring.h>  
#include <stdio.h> 
#include <sstmac/common/driver_util.h>
#include <sstmac/common/factories/topology_factory.h>

using std::set;
using std::vector;
using std::string;

namespace sstmac
{
  namespace hw
  {
    SSTRegisterTopology("torus | hdtorus", hdtorus, "hdtorus implements a high-dimension torus with an arbitrary number of dimensions");
    hdtorus::ptr
    hdtorus::construct(sim_parameters::ptr params)
    {
      bool errorflag = false;
      std::vector<int> args = get_intvec(
          params->get_param("topology_geometry").c_str(), errorflag);
      int nps = params->get_optional_int_param("network_nodes_per_switch", 0);

      return ptr(new hdtorus(args, nps, 1));

    }

    // ------------------------------------------------------------------------------------------------------
    /*  Computes the index array for all vertices  for a high dimensional torus
     It uses dsize : the number of dimensions
     dimensions_:  an array of size dsize that stores each dimension
     */

    void
    hdtorus::compute_vertex_numbers(std::vector<std::vector<int> > &vertices,
        int P)
    {
      int bsize = P;

      int64_t dsize = dimensions_.size();
      /* Below bsize  denotes how many consecutive  processors will be numbered
       the same in the  d-tg digit.  Note that  the for loop with  index i
       does not necessarily run P  times, since i is incremented each time
       in the innermost loop
       */
      for (int d = 0; d < dsize; d++)
        {
          bsize = bsize / dimensions_[d];
          for (int i = 0; i < P;)
            for (int k = 0; k < dimensions_[d]; k++)
              for (int j = 0; j < bsize; j++)
                vertices[i++][d] = k;
        }
    }

    // ------------------------------------------------------------------------------------------------------
    // Generates  the vertex numbers --and the vertices-- for a given number
    // of processors, P
    //

    void
    hdtorus::compute_vertices(std::vector<std::vector<int> > &vertices,
        set<vertex> &vertices_set, int id)
    {
      int64_t i, P;
      int64_t dsize = dimensions_.size();

      if (!vertices.size())
        {
          std::cout << " empty vertices in compute_vertices \n";
          std::cout.flush();
          for (i = 0, P = 1; i < dsize; P *= dimensions_[i++])
            ;
        }
      P = vertices.size();

      if (id < 0) /* single processor at each node */
        for (int i = 0; i < P; i++)
          vertices_set.insert(vertex(vertices[i]));
      else
        /* multiple processors at each node */
        for (int i = 0; i < P; i++)
          {
            vertices[i][dsize] = id;
            vertices_set.insert(vertex(vertices[i]));
          }
    }

    // --------------------------------------------------------------------------------------------------------
    // Generates the edges

    void
    hdtorus::compute_switch_edges(std::vector<std::vector<int> > &vertices,
        set<edge> &edges)
    {
      int jump, round, k;

      int P = 1;
      int64_t dsize = dimensions_.size();

      for (int i = 0; i < dsize; i++)
        P *= dimensions_[i];
      if (!direct_)
        for (int i = 0; i < P; i++)
          vertices[i][dsize] = nndpersw_;

      // here jump denotes the difference between processor numbers  and we decrease  gradually to 1.
      // e.g. in a 2x3x4x5 torus  jump will start with 120/2=60, andconnect vertices 0--60, 1--61, ..
      // then it will be 60/3=20 and connect 0 --20 etc.
      jump = round = P;
      for (int i = 0; i < dsize; i++)
        {
          jump = jump / dimensions_[i];
          if (wrapped_[i])
            { /* Torus network with edges wrapped around */
              for (int j = 0; j < P; j++)
                {
                  k = j - jump;
                  if (!vertices[j][i]) /* wrap around edge from 0 to the top  */
                    k += round;
                  edges.insert(edge(vertices[j], vertices[k]));

                  k = j + jump;
                  if (vertices[j][i] == dimensions_[i] - 1) /* wrap around edge from the top to zero */
                    k -= round;
                  edges.insert(edge(vertices[j], vertices[k]));
                }
            }
          else
            { /* Mesh like network with no wrap around edges */
              for (int j = 0; j < P; j++)
                {
                  if (vertices[j][i] > 0)
                    edges.insert(edge(vertices[j], vertices[j - jump]));
                  if (vertices[j][i] < dimensions_[i] - 1)
                    edges.insert(edge(vertices[j], vertices[j + jump]));
                }
            }
          round = jump;
        }
    }

    // --------------------------------------------------------------------------------------------------------
    // Connects nodes and switches

    void
    hdtorus::compute_node_edges(std::vector<std::vector<int> > &vertices,
        set<edge> &edges, int nndpersw)
    {
      int64_t dsize = dimensions_.size();

      std::vector<int> vi(dsize + 1); // vi specifies a node

      for (int i = 0; i < n() / nndpersw; i++)
        {
          for (int j = 0; j < dsize; j++)
            vi[j] = vertices[i][j];
          vertices[i][dsize] = nndpersw;
          for (vi[dsize] = 0; vi[dsize] < nndpersw; vi[dsize]++)
            {
              edges.insert(edge(vi, vertices[i])); // node to switch
              edges.insert(edge(vertices[i], vi)); // switch to node
            }
        }
    }

    // --------------------------------------------------------------------------------------------------------
    /*  Generates a high dimensional torus network
     Inputs:
     dimensionarray: specifies the number of processors in each dimension
     and whether  the links are wrapped around  (torus) or not (mesh).
     dimensionarray[i] <0 refers to absence of  wrap around edges.
     dimensionarray[i] >0 refers to presence of wrap around edges.

     nndpersw is the number of nodes per switch.
     widtharrg specifies the width between the switches and the  nodes.
     */
    hdtorus::hdtorus(vector<int> dimensionarray, int nndpersw, int widtharg)
    {
      int P, i;

      // Pre-process  the input;
      int64_t dsize = dimensionarray.size();
      std::vector<std::vector<int> > vertices;

      dimensions_.resize(dsize);
      wrapped_.resize(dsize);

      for (i = 0; i < dsize; i++)
        {
          if (dimensionarray[i] == 0)
            {
              throw illformederror(
                  "hdtorus::hdtorus: dimension of 0 not allowed");
            }
          else if (dimensionarray[i] < 0)
            {
              wrapped_[i] = 0;
              dimensions_[i] = -dimensionarray[i];
            }
          else if (dimensionarray[i] > 2)
            {
              dimensions_[i] = dimensionarray[i];
              wrapped_[i] = 1;
            }
          else
            {
              // This handles the case of dimension = 1 or 2, where
              // wrapping is turned off.
              dimensions_[i] = dimensionarray[i];
              wrapped_[i] = 0;
            }
        }

      direct_ = (nndpersw <= 0);
      nndpersw_ = nndpersw;
      width = widtharg;

      // Generate the  nodes.
      for (i = 0, P = 1; i < dsize; P *= dimensions_[i++])
        ;

      vertices.resize(P);
      if (direct_)
        for (i = 0; i < P; vertices[i++].resize(dsize))
          ;
      else
        for (i = 0; i < P; vertices[i++].resize(dsize + 1))
          ;

      compute_vertex_numbers(vertices, P);

      if (direct_)
        { // each node is connected to the  network directly.
          compute_vertices(vertices, nodes_);
          set<edge> edges;
          compute_switch_edges(vertices, edges);
          g_.init(nodes_, edges);
        }
      else
        { // each node is connected to the network  through a switch
          for (int i = 0; i < nndpersw; i++)
            compute_vertices(vertices, nodes_, i);

          std::set<vertex> all(nodes_);
          // insert the switch vertices
          compute_vertices(vertices, all, nndpersw);

          set<edge> edges;
          compute_switch_edges(vertices, edges);

          compute_node_edges(vertices, edges, nndpersw);

          g_.init(all, edges);
        }

      init();
      if (!direct_)
        init_node_to_switch();
    }

    // --------------------------------------------------------------------------------------------------------

    /// Returns the diameter of the interconnect.
    int64_t
    hdtorus::diameter() const
    {
      int D, i;
      int64_t dsize = dimensions_.size();

      for (D = i = 0; i < dsize; i++)
        if (wrapped_[i])
          D += (dimensions_[i] + 1) / 2;
        else
          D += dimensions_[i];
      return (D);
    }

    // --------------------------------------------------------------------------------------------------------

    /// Returns the average distance to all nodes.
    /* double hdtorus::a() const {
     std::cout << "Average distance in HD trous is not implemented yet ";
     return 0.0;
     }*/

    // --------------------------------------------------------------------------------------------------------

    /// Returns the minimum number of edges sliced by bisection.

    int64_t
    hdtorus::b() const
    {
      int x;
      int64_t dsize = dimensions_.size();
      int mymin = dimensions_[0];

      if (wrapped_[0])
        mymin *= 2;
      for (int i = 1; i < dsize; i++)
        {
          x = dimensions_[i];
          if (wrapped_[i])
            x *= 2;
          if (mymin < x)
            mymin = x;
        }
      return (mymin);
    }
    // --------------------------------------------------------------------------------------------------------
    // The following is an internal function that is used when computing a route between two vertices.
    // It computes the difference  along with the direction to travel the shortest distance while
    // taking into account the wrapped edges.  <0 means route by decreasing the  processors number
    // (e.g., go north) and >0 means  increase the processor number (e.g., go south)

    void
    hdtorus::compute_diffvect(const vertex &v1, const vertex &v2,
        vector<int> &diffvect) const
    {
      int d1;
      int64_t dsize = dimensions_.size();

      for (int64_t i = 0; i < dsize; i++)
        {
          d1 = v2.v()[i] - v1.v()[i];

          if (d1 < 0)
            {
              if (wrapped_[i] && dimensions_[i] < -2 * d1)
                diffvect[i] = dimensions_[i] + d1;
              else
                diffvect[i] = d1;
            }
          else if (d1 > 0)
            {
              if (wrapped_[i] && dimensions_[i] < 2 * d1)
                diffvect[i] = -dimensions_[i] + d1;
              else
                diffvect[i] = d1;
            }
          else
            diffvect[i] = 0;
        }
    }

    // --------------------------------------------------------------------------------------------------------

    int
    hdtorus::shortest_path_length(const vertex &v1, const vertex &v2, int id) const
    {
      int D, i;
      int64_t dsize = dimensions_.size();

      vector<int> diffvect(dsize);
      if (!id)
        compute_diffvect(v1, v2, diffvect);

      for (D = i = 0; i < dsize; i++)
        D += abs(diffvect[i]);
      return (D);
    }

    // --------------------------------------------------------------------------------------------------------
    /** Fills in the walk for the route between two node vertices.  The
     default implementation uses shortest path routing. */
    // virtual void route(const vertex &v1, const vertex &v2, walk &w) const;

    void
    hdtorus::shortest_path(const vertex&v1, const vertex&v2, walk&w) const
    {
      int i, j;
      int64_t dsize = dimensions_.size();
      bool print_route = 0;
      vector<vertex_index> vertices;
      vector<int> v(v1.v().size());
      vector<int> diffvect(dsize);
      vector<lwedge> nxt;

      compute_diffvect(v1, v2, diffvect);

      if (v1 == v2)
        {
          w.init(vertices);
          return;
        }
      if (print_route)
        {
          printf(" The path is ");
          for (i = 0; i < v1.v().size(); i++)
            printf("%d ", v1.v()[i]);
        }

      vertices.push_back(g().vertex_to_index(vertex(v1)));
      for (i = 0; i < v.size(); ++i)
        v[i] = v1.v()[i];

      if (!direct_)
        {
          // if this is a node, we must first get to a switch
          if (v[dsize] < nndpersw_)
            {
              v[dsize] = nndpersw_;
              if (print_route)
                {
                  printf(" ~~> ");
                  for (i = 0; i < v.size(); i++)
                    printf("%d ", v[i]);
                }
              vertices.push_back(g().vertex_to_index(vertex(v)));

            }
          else
            std::cerr << " Something wrong:message starts at a switch \n";
        }

      for (i = 0; i < dsize; i++)
        {
          if (diffvect[i] < 0)
            for (j = diffvect[i]; j < 0; j++)
              {
                if (v[i])
                  v[i]--;
                else
                  v[i] = dimensions_[i] - 1;
                if (print_route)
                  {
                    printf("--> ");
                    for (int k = 0; k < v.size(); k++)
                      printf("%d ", v[k]);
                  }
                vertices.push_back(g().vertex_to_index(v));

              }
          else
            for (j = 0; j < diffvect[i]; j++)
              {
                v[i]++;
                if (v[i] == dimensions_[i])
                  v[i] = 0;
                if (print_route)
                  {
                    printf("--> ");
                    for (int k = 0; k < v.size(); k++)
                      printf("%d ", v[k]);
                  }

                vertices.push_back(g().vertex_to_index(v));

              }
        }
      if (!direct_)
        {
          if (print_route)
            {
              printf("~~> ");
              for (i = 0; i < v2.v().size(); i++)
                printf("%d ", v2.v()[i]);
            }
          vertices.push_back(g().vertex_to_index(vertex(v2)));
        }
      w.init(vertices);
      if (print_route)
        printf("\n");
    }

    /** Returns all edges that take us one hop from the current
     location and closer to the destination.
     @param loc The vertex for the current location.
     @param dst The destination vertex.
     @param nxt A vector of hops.
     The default implementation throws.
     */
    void
    hdtorus::next_hop(vertex_index loc, vertex_index dst,
        std::vector<lwedge> &nxt) const
    {
      int i, t;
      int64_t dsize = dimensions_.size();
      nxt.clear();

      if (loc == dst)
        {
          return;
        }

      const vertex &locv = g().index_to_vertex(loc);
      const vertex &dstv = g().index_to_vertex(dst);

      vertex hop(g().word_length());
      for (int i = 0; i < dsize; ++i)
        hop.v()[i] = locv.v()[i];

      // If this is not direct, handle the special cases of hops to or
      // from a node.
      if (!direct_)
        {
          if (locv.v()[dsize] < nndpersw_)
            {
              // loc is a node; next hop is a switch

              hop.v()[dsize] = nndpersw_;
              nxt.push_back(g().generate_lwedge(loc, hop));
              return;
            }
          for (i = 0; (i < dsize) && (locv.v()[i] == dstv.v()[i]); i++)
            ;
          if (i == dsize)
            {
              // ready to hop from a switch to the destination node
              nxt.push_back(g().generate_lwedge(loc, dst));
              return;
            }
          hop[dsize] = nndpersw_;
        }
      // direct routing  or not a  speciel case of indirect routing
      for (i = 0; i < dsize; i++)
        {
          bool flag = true;
          int a = locv.v()[i];
          int b = dstv.v()[i];
          int diff = b - a;
          if (diff)
            {
              int nx = dimensions_[i];
              if (wrapped_[i])
                {
                  if (nx == 2 * diff || nx == -2 * diff)
                    {
                      // Either way gets us closer in this case.
                      flag = false;
                      t = hop[i];
                      hop[i] = (a + 1) % nx;
                      nxt.push_back(g().generate_lwedge(loc, hop));

                      hop[i] = (a + nx - 1) % nx;
                      nxt.push_back(g().generate_lwedge(loc, hop));
                      hop[i] = t;
                    }

                  if (diff > 0 && 2 * diff > nx)
                    diff = diff - nx;
                  else if (diff < 0 && 2 * diff < -nx)
                    diff = diff + nx;
                }

              if (flag)
                {
                  // this routing applies to both mesh and torus routing
                  t = hop[i];
                  if (diff > 0)
                    hop[i] = (a + 1) % nx;
                  else
                    hop[i] = (a + nx - 1) % nx;

                  nxt.push_back(g().generate_lwedge(loc, hop));
                  hop[i] = t;
                }
            }
        }
    }

    // ----------------------------------------------------------------------------------

    /// Returns a descriptive name of the graph.
    string
    hdtorus::name() const
    {
      int64_t dsize = dimensions_.size();
      std::ostringstream ostr;
      ostr << "hdtorus(";
      for (int i = 0; i < dsize; i++)
        ostr << dimensions_[i] << ",";
      ostr << ")";
      return ostr.str();
    }

    std::vector<int>
    hdtorus::get_coords(long uid)
    {
      long div = 1;
      std::vector<int> ret(dimensions_.size());
      for (int i = 0; i < dimensions_.size(); i++)
        {
          ret[i] = (uid / div) % dimensions_[i];
          div *= dimensions_[i];
        }
      return ret;
    }

    long
    hdtorus::get_uid(std::vector<int> coords)
    {
      long ret = 0;
      long mult = 1;
      for (int i = 0; i < dimensions_.size(); i++)
        {
          ret += coords[i] * mult;
          mult *= dimensions_[i];
        }
      return ret;
    }

    std::string
    hdtorus::coord_string(std::vector<int> coords)
    {
      std::stringstream ss;
      ss << "(";
      for (int i = 0; i < dimensions_.size(); i++)
        {
          ss << coords[i] << ",";
        }
      ss << ")";
      return ss.str();
    }

    int64_t
    hdtorus::topology_size() const
    {
      int64_t ret = 1;
      for (int i = 0; i < dimensions_.size(); i++)
        {
          ret *= dimensions_[i];
        }
      return ret;
    }

    void
    hdtorus::connect_switches(switchtopology::maptype &switches,
        sim_parameters::ptr params, std::string linkbwname)
    {

      double linkbw = get_bandwidth(params->get_param(linkbwname).c_str());
      SSTMAC_DEBUG << "hdtorus: link bandwidth is " << linkbw << "\n";

      bool errorflag = false;

      std::vector<int> red;
      if (params->has_param("topology_redundant"))
        {
          red = get_intvec(params->get_param("topology_redundant").c_str(),
              errorflag);
        }
      else
        {
          for (int i = 0; i < dimensions_.size(); i++)
            {
              red.push_back(1);
            }
        }
      for (long i = 0; i < switches.size(); i++)
        {
          nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);
          std::vector<int> coords = get_coords(i);

          for (int z = 0; z < dimensions_.size(); z++)
            {
              std::vector<int> connect1 = coords;
              std::vector<int> connect2 = coords;
              if (coords[z] == 0)
                {
                  connect1[z]++;
                  connect2[z] = dimensions_[z] - 1;
                }
              else if (coords[z] == (dimensions_[z] - 1))
                {
                  connect1[z] = 0;
                  connect2[z]--;
                }
              else
                {
                  connect1[z]++;
                  connect2[z]--;
                }
              long uid1 = get_uid(connect1);
              long uid2 = get_uid(connect2);

              nodeaddress::ptr addr1 = nodeaddress::convert_from_uniqueid(uid1);
              nodeaddress::ptr addr2 = nodeaddress::convert_from_uniqueid(uid2);

              if (switches.find(addr1) == switches.end())
                {
                  std::cout << "uid: " << uid1 << ", addr: "
                      << addr1->to_string() << "\n";
                  throw ssterror(
                      "hdtorus::connect_switches() -- can't find switch 1 when connecting");
                }
              if (switches.find(addr2) == switches.end())
                {
                  std::cout << "uid: " << uid2 << ", addr: "
                      << addr2->to_string() << "\n";
                  throw ssterror(
                      "hdtorus::connect_switches() -- can't find switch 2 when connecting");
                }

              SSTMAC_DEBUG << "switch " << me->to_string() << " with coords "
                    << coord_string(coords) << " for dim " << z
                    << " connecting in the pos dir to " << coord_string(
                    connect1) << " whic is addr " << addr1->to_string() << "\n";
              SSTMAC_DEBUG << "switch " << me->to_string() << " with coords "
                    << coord_string(coords) << " for dim " << z
                    << " connecting in the neg dir to " << coord_string(
                    connect2) << " whic is addr " << addr2->to_string() << "\n";

              switches[me]->connect(z, router::POS, switches[addr1], linkbw,
                  red[z]);
              switches[me]->connect(z, router::NEG, switches[addr2], linkbw,
                  red[z]);
            }
        }

    }

    router::ptr
    hdtorus::create_router(nodeaddress::ptr addr, sim_parameters::ptr params)
    {
      int nps = std::max(nndpersw_, 1);
      router::ptr ret = torusrouter::construct(addr, nps, dimensions_);

      return ret;
    }

    void
    hdtorus::display_nodes(vismaptype &switches, vis::vis_engine::ptr eng,
        std::list<vis::vis_obj::ptr> &objs)
    {
      for (long i = 0; i < switches.size(); i++)
        {
          nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);

          std::vector<int> coords = get_coords(i);

          double x = 0, y = 0, z = 0;
          x = coords[0] * 2;
          if (coords.size() > 1)
            y = coords[1] * 2;

          if (coords.size() > 2)
            z = coords[2] * 2;

          vis::vis_obj::ptr vo = eng->create_cube(x, y, z, 1);
          switches[me]->set_vis_obj(vo);
          switches[me]->set_vis_engine(eng);
          objs.push_back(vo);

          for (int z = 0; z < dimensions_.size(); z++)
            {
              double x1 = coords[0] * 2;
              double y1 = coords[1] * 2;
              double z1 = coords[2] * 2;

              double x2 = coords[0] * 2;
              double y2 = coords[1] * 2;
              double z2 = coords[2] * 2;
              int dir;

              bool onegood = (coords[z] != dimensions_[z] - 1);
              bool twogood = (coords[z] != 0);

              if (z == 0)
                {

                  x1 += 1.5;
                  z1 += 0.5;
                  y1 += 0.25;

                  x2 -= 0.5;
                  z2 += 0.5;
                  y2 += 0.75;

                  dir = vis::vis_engine::VIS_Z;
                }
              else if (z == 1)
                {
                  y1 += 1.5;
                  z1 += 0.25;
                  x1 += 0.5;

                  y2 -= 0.5;
                  z2 += 0.75;
                  x2 += 0.5;
                  dir = vis::vis_engine::VIS_Y;
                }
              else
                {
                  z1 += 1.5;
                  y1 += 0.25;
                  x1 += 0.5;

                  z2 -= 0.5;
                  y2 += 0.75;
                  x2 += 0.5;
                  dir = vis::vis_engine::VIS_X;
                }

              if (onegood)
                {
                  vis::vis_obj::ptr c1 = eng->create_cylinder(x1, y1, z1,
                      0.125, 1, dir);
                  objs.push_back(c1);
                  switches[me]->set_link_vis(c1, z, router::POS);
                }

              if (twogood)
                {
                  vis::vis_obj::ptr c2 = eng->create_cylinder(x2, y2, z2,
                      0.125, 1, dir);
                  switches[me]->set_link_vis(c2, z, router::NEG);
                  objs.push_back(c2);
                }
            }
        }
    }

  }
} //end of namespace sstmac


