/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
// torus.cc: Implementation of torus networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>
#include <sstream>
#include <sstmac/hardware/network/topology/oldtorus.h>
#include <sstmac/hardware/network/topology/ring.h>
#include <sstmac/common/driver_util.h>
#include <sstmac/common/factories/topology_factory.h>


using std::string;

namespace sstmac
{
  namespace hw
  {
    SSTRegisterTopology("oldtorus | torus_2D | torus_3D | torus3D | torus2D", torus);

    torus::ptr
    torus::construct(sim_parameters::ptr params)
    {
      bool errorflag = false;
      std::vector<int> args = get_intvec(
          params->get_param("topology_geometry").c_str(), errorflag);
      int nps = params->get_optional_int_param("network_nodes_per_switch", 0);
      std::string name = params->get_param("topology_name");
      if (name == "torus_2D" || name == "torus2D" || args.size() == 2)
        {
          int wid = 1;
          if (args.size() == 3)
            {
              nps = args[2];
            }else if(args.size() == 4){
                nps = args[2];
                wid = args[3];
            }
          return construct_2D(args[0], args[1], nps, wid);
        }
      else if (args.size() == 3)
        {
          return construct_3D(args[0], args[1], args[2], nps, 1);
        }
      else if (args.size() == 4)
        {
          return construct_3D(args[0], args[1], args[2], args[3], 1);
        }
      else if (args.size() == 5)
        {
          return construct_3D(args[0], args[1], args[2], args[3], args[4]);
        }
      else
        {
          throw ssterror("oldtorus:: needs 2 or 3 parameters");
        }
    }

    torus::torus()
    {
    }

    void
    torus::init_torus_2D(int x, int y, int nndpersw, int widtharg)
    {
      dimensions_.resize(2);
      dimensions_.at(0) = x;
      dimensions_.at(1) = y;
      width = widtharg;
      direct_ = (nndpersw == 0);
      nndpersw_ = nndpersw;

      topology::ptr n1 = ring::construct(x, nndpersw);
      topology::ptr n2 = ring::construct(y, (nndpersw > 0 ? 1 : 0));
      assemble(n1, n2);

      std::ostringstream ostr;
      ostr << "Torus" << "(" << x << "," << y << ",nps=" << nndpersw << ")";
      name_ = ostr.str();

      SSTMAC_DEBUG << "Initialized " << name_ << ", vertexes have "
            << nodes_.begin()->v().size() << " indices \n";
    }

    void
    torus::init_torus_3D(int x, int y, int z, int nndpersw, int widtharg)
    {
      dimensions_.resize(3);
      dimensions_.at(0) = x;
      dimensions_.at(1) = y;
      dimensions_.at(2) = z;
      width = widtharg;
      direct_ = (nndpersw == 0);
      nndpersw_ = nndpersw;

      topology::ptr n1 = torus::construct_2D(x, y, nndpersw);
      topology::ptr n2 = ring::construct(z, (nndpersw > 0 ? 1 : 0));
      assemble(n1, n2);

      std::ostringstream ostr;
      ostr << "Torus" << "(" << x << "," << y << "," << z << ",nps="
          << nndpersw << ")";
      name_ = ostr.str();

      SSTMAC_DEBUG << "Initialized " << name_ << ", vertexes have "
            << nodes_.begin()->v().size() << " indices \n";
    }

    torus::ptr
    torus::construct_2D(int n1, int n2, int nndpersw, int widtharg)
    {
      torus::ptr t(new torus);
      t->init_torus_2D(n1, n2, nndpersw, widtharg);
      return t;
    }

    torus::ptr
    torus::construct_3D(int n1, int n2, int n3, int nndpersw, int widtharg)
    {
      torus::ptr t(new torus);
      t->init_torus_3D(n1, n2, n3, nndpersw, widtharg);
      return t;
    }

    torus::~torus()
    {
    }

    std::string
    torus::name() const
    {
      return name_;
    }

    int
    torus::edge_width(edge_index e) const
    {
      const lwedge &lwe = g().index_to_edge(e);
      if (is_switch(lwe.from()) && is_switch(lwe.to()))
        return width;
      return 1;
    }

  }
} //end of namespace sstmac
