/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/nic/rdma_nic.h>
//#include <sstmac/simulator/networkdata.h>
//#include <sstmac/process/recvhandler.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/common/factories/nic_factory.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegisterNic("rdma | rdma_nic", rdma_nic, "implements a nic that is capable of sending rmda get/put without interrupting CPU");

    // Pointer types.

    using namespace sstmac::hw;

    rdma_nic::rdma_nic() :
          networkinterface()
    {
    }

    rdma_nic::ptr
    rdma_nic::construct(const sim_parameters::ptr& params)
    {
      return rdma_nic::ptr(new rdma_nic);
    }

    //
    // Goodbye.
    //
    rdma_nic::~rdma_nic() throw ()
    {
    }

    networkinterface::ptr
    rdma_nic::clone(const node::ptr &parent,
        const sstmac::hw::interconnect::ptr &interconn,
        const sim_parameters::ptr& params) const
    {
      rdma_nic::ptr ret = rdma_nic::construct( params);
      ret->init(parent, interconn);
      return ret;
    }

    void
    rdma_nic::handle(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "rdma_nic: " << get_parent_id()->to_string()
            << "handle message: " << msg->to_string() << "\n\n";
      switch (msg->get_type())
        {
      case sst_message::RDMA_GET_REQ:
        send_rdma_rsp(msg);
        break;
      case sst_message::RDMA_GET_RSP:
        recv_rdma_rsp(msg);
        break;
      case sst_message::RDMA_PUT_REQ:
      case sst_message::RDMA_PUT_RSP:
        throw unimplementederror("rdma_nic: RDMA PUT not yet allowed");
        break;
      case sst_message::SEND_ACK:
        schedule(now(), parent_, msg);
        break;
      default:
        recv(msg);
        break;
        }
    }

    void
    rdma_nic::set_eventmanager(const eventmanager::ptr &m)
    {
      networkinterface::set_eventmanager(m);
    }

    void
    rdma_nic::send(const nodeaddress::ptr &recver,
        const sst_message::ptr &payload)
    {
      use_me();

      eventcallback::ptr cb = new_callback(rdma_nic::ptr(this),
          &rdma_nic::send_to_network_link, payload, recver);

      lib_mem_->nic_read(payload->get_byte_length(), payload->source_task(), cb);
    }

    //
    // Schedule a send operation to happen at the given time.
    //
    void
    rdma_nic::send_to_network_link(const sst_message::ptr& payload,
        const nodeaddress::ptr& recver)
    {
      SSTMAC_DEBUG << "rdma_nic::send(" << recver->to_string() << ", "
            << payload->to_string() << ")\n";

      if (payload == 0)
        {
          std::cout << "I'm here throwing this payload error" << std::endl;
          throw nullerror("simplenic::send: null payload pointer.");
        }

      timestamp delayed = now(); //don't model delay for now

      SSTMAC_DEBUG << "rdma_nic:send " << get_parent_id()->to_string()
            << "- sending now \n\n";

      interconn_->send(delayed, payload);

      if (spy_)
        {
          spy_->add_one(payload->fromaddr_->unique_id(),
              payload->toaddr_->unique_id());
          spy_bytes_->add(payload->fromaddr_->unique_id(),
              payload->toaddr_->unique_id(), payload->get_byte_length());
        }
    }

    void
    rdma_nic::recv_rdma_rsp(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "rdma_nic::recv_rdma_rsp: "
            << get_parent_id()->to_string()
            << "\n\n";
      //perform the memory access
      eventcallback::ptr cb = new_callback(rdma_nic::ptr(this),
          &rdma_nic::notify_session_manager, msg);

      //parent_->get_lib_mem()->rdma_write(msg->get_byte_length(), msg->dest_task(), cb);
      lib_mem_->rdma_put(msg->get_byte_length(), msg->dest_task(), cb);
    }

    void
    rdma_nic::send_rdma_rsp(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "rdma_nic::send_rdma_rsp: "
            << get_parent_id()->to_string()
            << " " << msg->to_string() << "\n\n";
      /** pull the data in from memory and fire off to the requesting node */
      eventcallback::ptr cb = new_callback(rdma_nic::ptr(this),
          &rdma_nic::finalize_rdma_rsp, msg);

      msg->mutate(sst_message::RDMA_GET_REQ_TO_RSP);

      lib_mem_->rdma_get(msg->get_byte_length(), msg->source_task(), cb);
    }

    void
    rdma_nic::finalize_rdma_rsp(const sst_message::ptr& msg)
    {
      SSTMAC_DEBUG << "rdma_nic::finalize_rdma_rsp: "
            << get_parent_id()->to_string()
            << "\n\n";
      if (!msg->fromaddr_->equals(get_parent_id()))
        throw illformederror(
            "rdma_engine: from address in get request improperly formatted");

      send_to_network_link(msg, msg->toaddr_);
    }

    void
    rdma_nic::recv(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "rdma_nic::recv: "
            << get_parent_id()->to_string()
            << " "
            << msg->to_string() << "\n\n";

      eventcallback::ptr cb = new_callback(rdma_nic::ptr(this),
          &rdma_nic::notify_session_manager, msg);
      lib_mem_->nic_write(msg->get_byte_length(), msg->dest_task(), cb);
    }

    void
    rdma_nic::notify_session_manager(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "rdma_nic::notify_session_manager: "
            << get_parent_id()->to_string()
            << " " << msg->to_string() << "\n\n";
      get_parent_node()->handle(msg);
    }

    void
    rdma_nic::validate_msg_supported(const sst_message::ptr &msg) const
    {
      switch (msg->get_type())
        {
      case sst_message::MPI:
      case sst_message::HPX:
      case sst_message::SEND_ACK:
      case sst_message::ACTOR:
      case sst_message::NETWORK:
      case sst_message::RDMA_GET_REQ:
      case sst_message::RDMA_GET_RSP:
        return; //valid
        }
      sst_throw(unimplementederror, "%s: network interface card received invalid message type %s",
                to_string().c_str(), sst_message::tostr(msg->get_type()));
    }

  }
} // end of namespace sstmac.
