/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_COMPONENTS_NODE_NODE_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_COMPONENTS_NODE_NODE_H_INCLUDED

#include <sstmac/common/timestamp.h>

#include <sstmac/common/ptr_type.h>

#include <sstmac/common/eventmodule.h>
#include <sstmac/software/ami/ami.h>
#include <sstmac/common/debug.h>

//#include <sstmac/software/services/launch/launcher.h>

#include <sstmac/software/services/launch/allocation/allocationstrategy.h>

#include <sstmac/hardware/failure/failable.h>

#include <list>
#include <string>

namespace sstmac
{
  namespace sw
  {
    class operatingsystem;
    class launcher;
    class environment;
    class service;
  }

  namespace hw
  {

    // Forward declarations of relevant types.

    class networkinterface;

    /**
     * An interface for "physical" nodes -- may eventually be decorated
     * with additional functionality (OS/cpu modeling or whatever).
     */
    class node : public eventmodule, public sstmac::sw::AMI, public failable
    {

      friend class message_handler;

    public:
      typedef boost::intrusive_ptr<node> ptr;
      typedef node* parent_ptr;

    protected:
      class message_handler : public eventhandler
      {
      protected:
        node::ptr parent_;

        message_handler(const node::ptr &p) :
          parent_(p)
        {
        }

      public:
        virtual void
        handle(const sst_message::ptr &msg);

        static boost::intrusive_ptr<message_handler>
        construct(const node::ptr &p)
        {
          boost::intrusive_ptr<message_handler> ret(new message_handler(p));
          return ret;
        }

        virtual std::string
        to_string() const
        {
          return "node::message_handler";
        }
      };

      virtual bool
      has_failed();
      virtual void
      reset_failed();

      boost::intrusive_ptr<message_handler> msg_handler_;

      boost::intrusive_ptr<sstmac::sw::environment> env_;

      boost::intrusive_ptr<sw::operatingsystem> os_;

      /// The id for this node.
      nodeaddress::ptr nodeid_;

      boost::intrusive_ptr<sstmac::sw::launcher> launcher_;

      static debug dbg_;

      sim_parameters::ptr params_;

      node(const sim_parameters::ptr &params, nodeaddress::ptr my_id);

      virtual
      ~node();

      static sstmac::sw::allocationstrategy::ptr allocation_;

      boost::intrusive_ptr<networkinterface> nic_;
    public:

      /// Get a new node of the same type as the current one.
      virtual node::ptr
      clone(nodeaddress::ptr my_id) const = 0;

      /// Assign a node index and event call interface for this node.
      /// \throw nullerror if ev is null.
      virtual void
      initialize(eventmanager::ptr man);

      void set_nic(const boost::intrusive_ptr<networkinterface> &n);

      boost::intrusive_ptr<networkinterface>
      get_nic() const;

      boost::intrusive_ptr<sw::operatingsystem>
      get_os() const;

      virtual void
      handle(const sst_message::ptr &msg);

      virtual std::string
      to_string() const
      {
        return "node";
      }

      virtual void finish();

      virtual long
      failure_id() const
      {
        nodeid_->unique_id();
      }

      nodeaddress::ptr
      get_id()
      {
        return nodeid_;
      }

      virtual std::list<failable::ptr>
      get_failure_components() = 0;

      void
      add_service(boost::intrusive_ptr<sstmac::sw::service> serv);

      //virtual void
      //init_failure_model(double l_proc, double l_mem, double l_disk, double l_nic, failable::ptr ps, double p_ps_node) = 0;

    };

  }
} // end of namespace sstmac

#endif
