/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/simpleprocessor-config.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>
#include <sstmac/hardware/memory/simple_memorymodel.h>
#include <sstmac/hardware/memory/numa_memorymodel.h>
#include <sstmac/hardware/node/simplenode.h>
#include <sstmac/hardware/processor/migratingcomputescheduler.h>
#include <sstmac/hardware/nic/networkinterface.h>
#include <sstmac/software/process/operatingsystem.h>

#include <sstmac/software/libraries/filesystem/disk_message.h>
#include <sstmac/software/libraries/filesystem/nppfs_server.h>

#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/machine_descriptor.h>
#include <iostream>
#include <sstream>
#include <sstmac/software/services/launch/launcher.h>
#include <sstmac/common/factories/node_factory.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegisterNode("simple | simplenode", simplenode, "simple node which implements basic OS/compute scheduling functionality");
    using namespace sstmac::sw;

    simplenode::simplenode(const sim_parameters::ptr &params,
        nodeaddress::ptr my_id) :
      node(params, my_id)
    {
      os_->setAMI(sstmac::sw::AMI::AMI_COMM, this);
      os_->setAMI(sstmac::sw::AMI::AMI_COMPUTE, this);
      os_->setAMI(sstmac::sw::AMI::AMI_QUERY, this);
      os_->setAMI(sstmac::sw::AMI::AMI_HW, this);

      ncores_ = params->get_int_param("node_cores");
      nsockets_ = params->get_optional_int_param("node_sockets", 1);
      memory_latency_ = params->get_time_param("node_mem_latency");
      bool errorflag = false;
      memory_bandwidth_ = get_bandwidth(
          params->get_param("node_mem_bandwidth").c_str(), errorflag);
      freq_ = params->get_double_param("node_frequency");

      timestamp pre(-1);
      if (params->has_param("node_preemption"))
        {
          pre = params->get_time_param("node_preemption");
        }
      preempt_ = pre;

      SSTMAC_DEBUG << "constructed a simplenode with parameters mem_lat = "
            << memory_latency_ << ", memory_bandwidth = " << memory_bandwidth_
            << ", freq = " << freq_ << ", cores = " << ncores_ << "\n";
    }

    void
    simplenode::initialize(eventmanager::ptr man)
    {

      node::initialize(man);

      SSTMAC_DEBUG << "starting simplenode initialization \n";

      double tcycle = 1.0 / (freq_);
      double mop_freq = params_->get_optional_double_param(
          "node_memop_frequency", freq_);
      double memop_cycle = 1.0 / mop_freq;

      sstmac::hw::simpleprocessor_config::ptr cfg =
          sstmac::hw::simpleprocessor_config::construct(ncores_, freq_, 10, 30);

      compute_sched_ = sstmac::hw::migratingcomputescheduler::construct(
          nodeid_->unique_id(), ncores_, nsockets_, preempt_, os_);
      compute_sched_->set_eventmanager(man);
      proc_model_ = sstmac::hw::simpleprocessor::construct(cfg, msg_handler_,
          compute_sched_);

      sstmac::hw::instructionprocessor_config::ptr pcfg =
          sstmac::hw::instructionprocessor_config::construct(ncores_,
              eventdata::INST_FLOP, tcycle,
              eventdata::INST_MEMOP, memop_cycle,
              eventdata::INST_MEMOP_LOCAL, memop_cycle,
              eventdata::INST_MEMOP_RAND, memop_cycle);

      proc_inst_ = sstmac::hw::instructionprocessor::construct(pcfg,
          msg_handler_, compute_sched_);
      if (params_->has_param("node_frequency_stdev"))
        {
          double maxfreq = freq_ + params_->get_double_param(
              "node_frequency_stdev");
          double tmin = 1.0 / maxfreq;
          double delta_t = tcycle - tmin;
          proc_inst_->set_flop_distribution(delta_t);
          proc_inst_->set_memop_distribution(delta_t);
        }

      std::string mem_model = params_->get_optional_param("node_memory_model",
          "simple");
      if (mem_model == "simple")
        {
          mem_model_ = sstmac::hw::simple_memorymodel::construct(
              ptr(this),
              memory_bandwidth_, memory_latency_, msg_handler_,
              nodeid_->unique_id());
        }
      else if (mem_model == "numa")
        {
          double bus_bw = get_bandwidth(params_->get_param("numa_bus_bandwidth").c_str());
          int max_nsegments = params_->get_optional_int_param("numa_flow_max_nsegments", flow::default_max_nsegments);
          long min_nbytes_per_segment = params_->get_optional_long_param("numa_flow_min_nybtes_per_segmenet", flow::default_min_bytes_per_segment);
          double nic_bw = get_bandwidth(params_->get_param("nic_mem_link_bandwidth").c_str());
          timestamp nic_lat = params_->get_time_param("nic_mem_link_latency");
          numa_memorymodel::ptr numa_model = sstmac::hw::numa_memorymodel::construct(
                                                 ptr(this), msg_handler_,
                                                 memory_latency_, bus_bw, memory_bandwidth_,
                                                 nic_lat, nic_bw,
                                                 nsockets_, nodeid_->unique_id(),
                                                 max_nsegments, min_nbytes_per_segment,
                                                 compute_sched_);

          char param_name[100];
          //for now - I don't know a better way to do this - loop all possible connections
          for (int i=0; i < nsockets_; ++i){
              for (int j=0; j < nsockets_; ++j){
                  if (i==j)
                      continue;

                  sprintf(param_name, "numa_hypertransport_bandwidth_%d_%d", i, j);
                  if (params_->has_param(param_name)){
                      double link_bw = get_bandwidth(params_->get_param(param_name).c_str());
                      numa_model->add_hypertransport_link(i, j, link_bw);
                  }
              }
          }
          mem_model_ = numa_model;
        }
      else
        {
          std::string errormsg = "simplenode: invalid memory model "
              + mem_model;
          throw inputerror(errormsg);
        }
      mem_model_->set_eventmanager(man);

      if (params_ && params_->has_param("disk_type"))
        {
          if (params_->get_param("disk_type") == "vesper")
            {
              disk_ = sstmac::hw::disk_controller::construct(
                  nodeid_->unique_id(), params_, msg_handler_, man);
            }
        }

      if (params_ && params_->has_param("node_nppfs_server"))
        {
          long nid = params_->get_long_param("node_nppfs_server");
          nodeaddress::ptr addr = nodeaddress::convert_from_uniqueid(nid);
          nppfs_server_ = addr;

          std::string fname = params_->get_param("node_nppfs_diskfile");
          sstmac::sw::nppfs_server::ptr serv =
              sstmac::sw::nppfs_server::construct(fname, addr->equals(nodeid_));
          os_->register_ser(serv);
          serv->start();

        }

      SSTMAC_DEBUG << "simplenode initialization done, starting things \n";
      os_->start_apps();

      launcher_->start();
    }

    node::ptr
    simplenode::clone(nodeaddress::ptr my_id) const
    {
      node::ptr ret = construct(params_, my_id);

      return ret;
    }

    void
    simplenode::finish()
    {
        if (mem_model_){
            mem_model_->finish();
            mem_model_ = 0;
        }
        if (proc_model_)
            proc_model_ = 0;
        if (proc_inst_)
            proc_inst_ = 0;
        if (compute_sched_)
            compute_sched_ = 0;
        if (disk_)
            disk_ = 0;

        node::finish();
    }

    void
    simplenode::execute_kernel(AMI::AMI_COMM_FUNC func, const sst_message::ptr& data)
    {
        bool failed = false;

        switch(func)
        {
        case sstmac::sw::AMI::AMI_COMM_SEND:
          {

            if (data->get_type() == sst_message::FILEMETA)
              {
                data->toaddr_ = nppfs_server_;

                if (!nppfs_server_)
                  throw ssterror(
                      "simplenode: I don't have an address for the nppfs server");
              }

            nodeaddress::ptr nid = boost::dynamic_pointer_cast<nodeaddress>(
                data->toaddr_);
            data->fromaddr_ = nodeid_;

            SSTMAC_DEBUG << "Node(" << nodeid_->to_string() << "): sending to "
                  << data->toaddr_->to_string() << "\n";

            if (data->toaddr_->equals(nodeid_))
              {
                schedule(now(), msg_handler_, data);
                // msg_handler_->handle(data);
              }
            else
              nic_->send(nid, data);

            failed = nic_->has_failed() || mem_model_->has_failed();
            break;
          }
        default:
          sst_throw(ssterror, "simplenode: cannot process kernel %s", AMI::tostr(func));
          break;
        }
    }

    void
    simplenode::execute_kernel(AMI::AMI_COMP_FUNC func, const sst_message::ptr& data)
    {
        bool failed = false;

        switch(func)
        {
        case sstmac::sw::AMI::AMI_COMP_TIME:
          {

            proc_model_->compute(data, os_->get_current_threadid());

            failed = compute_sched_->has_failed();

            break;
          }

        case sstmac::sw::AMI::AMI_COMP_INSTR_TIME:
        {
            data->set_event_time(proc_inst_->compute_time(data, os_->get_current_threadid()));
            break;
        }
        case sstmac::sw::AMI::AMI_COMP_INSTR:
          {

            proc_inst_->compute(data, os_->get_current_threadid());

            failed = compute_sched_->has_failed();

            break;
          }

        case sstmac::sw::AMI::AMI_COMP_DISKACCESS:
          {
            sstmac::sw::disk_message::ptr dm = boost::dynamic_pointer_cast<
                sstmac::sw::disk_message>(data);
            disk_->request_access(data);

            failed = disk_->has_failed();
            break;
          }

        case sstmac::sw::AMI::AMI_COMP_SCHEDULE:
          {
            compute_sched_->reserve(data, os_->get_current_threadid());
            break;
          }

        case sstmac::sw::AMI::AMI_COMP_RELEASE:
          {
            compute_sched_->release(data, os_->get_current_threadid());
            break;
          }

        case sstmac::sw::AMI::AMI_COMP_MEMREAD:
          {
            mem_model_->read(data);
            failed  = mem_model_->has_failed();
            break;
          }

        case sstmac::sw::AMI::AMI_COMP_MEMWRITE:
          {
            mem_model_->write(data);
            failed = mem_model_->has_failed();
            break;
          }
        case sstmac::sw::AMI::AMI_COMP_NIC_READ:
        {
            mem_model_->nic_read(data);
            failed = mem_model_->has_failed();
            break;
        }
        case sstmac::sw::AMI::AMI_COMP_NIC_WRITE:
        {
            mem_model_->nic_write(data);
            failed = mem_model_->has_failed();
            break;
        }
        case sstmac::sw::AMI::AMI_COMP_RDMA_GET:
        {
            mem_model_->rdma_get(data);
            failed = mem_model_->has_failed();
            break;
        }
        case sstmac::sw::AMI::AMI_COMP_RDMA_PUT:
        {
            mem_model_->rdma_put(data);
            failed = mem_model_->has_failed();
            break;
        }
        case sstmac::sw::AMI::AMI_COMP_SLEEP:
          {
            timestamp delay = now() + data->get_time();
            SSTMAC_DEBUG << "node (" << nodeid_->to_string()
                  << "): sleeping until " << delay << "\n";
            schedule(delay, ptr(this), data);
            break;
          }
        default:
          sst_throw(ssterror, "simplenode: cannot process kernel %s", AMI::tostr(func));
          break;
        }
    }

    bool
    simplenode::has_failed()
    {
      return nic_->has_failed() || mem_model_->has_failed()
          || compute_sched_->has_failed() || failable::has_failed();
    }
    void
    simplenode::reset_failed()
    {
      nic_->reset_failed();
      mem_model_->reset_failed();
      compute_sched_->reset_failed();
      failable::reset_failed();
    }

    ptr_type::ptr
    simplenode::execute_kernel(AMI::AMI_HW_FUNC func)
    {
        switch(func)
        {
        case sstmac::sw::AMI::AMI_CATASTROPHIC_INTERRUPT:
          {
            cancel_all_messages();
            mem_model_->cancel_all_messages();
            compute_sched_->cancel_all_messages();
            nic_->cancel_all_messages();
            return ptr_type::ptr();
            break;
          }
        default:
            sst_throw(valueerror, "simplenode: cannot process kernel %s", tostr(func));
        }
    }

    ptr_type::ptr
    simplenode::execute_kernel(AMI::AMI_QUERY_FUNC func)
    {
      switch (func)
        {
      case sstmac::sw::AMI::AMI_QUERY_TOPOLOGY:
        {
          sstmac::ptr_type::ptr top = nic_->get_topology();
          return top;
          break;
        }
      case sstmac::sw::AMI::AMI_QUERY_ADDRESS:
        {
          return nodeid_;
          break;
        }

      case sstmac::sw::AMI::AMI_QUERY_MACHINE:
        {
          return sstmac::sw::machine_descriptor::construct(
              nic_->get_topology()->n(), nic_->get_topology()->name());
          break;
        }

      default:
          sst_throw(valueerror, "simplenode: cannot process kernel %s", tostr(func));
        }
    }

    timestamp
    simplenode::now() const
    {
      return eventmodule::now();
    }

    bool
    simplenode::kernel_supported(AMI::AMI_COMP_FUNC func) const
    {
        switch (func)
        {
        case sstmac::sw::AMI::AMI_COMP_TIME:
        case sstmac::sw::AMI::AMI_COMP_SLEEP:
        case sstmac::sw::AMI::AMI_COMP_INSTR:
        case sstmac::sw::AMI::AMI_COMP_MEMREAD:
        case sstmac::sw::AMI::AMI_COMP_MEMWRITE:
          return true;
        case sstmac::sw::AMI::AMI_COMP_DISKACCESS:
          return disk_;
        }
        return false;
    }

    bool
    simplenode::kernel_supported(AMI::AMI_COMM_FUNC func) const
    {
        switch (func)
        {
        case sstmac::sw::AMI::AMI_COMM_SEND:
        case sstmac::sw::AMI::AMI_COMM_NIC_SEND_ACK:
          return true;
        }
        return false;
    }

    bool
    simplenode::kernel_supported(AMI::AMI_HW_FUNC func) const
    {
        switch (func)
        {
        case sstmac::sw::AMI::AMI_CATASTROPHIC_INTERRUPT:
          return true;
        }
        return false;
    }

    bool
    simplenode::kernel_supported(AMI::AMI_QUERY_FUNC func) const
    {
        switch (func)
          {
        case sstmac::sw::AMI::AMI_QUERY_TOPOLOGY:
        case sstmac::sw::AMI::AMI_QUERY_ADDRESS:
          return true;
          }
        return false;
    }

    std::list<failable::ptr>
    simplenode::get_failure_components()
    {
      std::list<failable::ptr> ret;
      if (compute_sched_)
        ret.push_back(compute_sched_);
      if (mem_model_)
        ret.push_back(mem_model_);
      if (disk_)
        ret.push_back(disk_);
      if (nic_)
        ret.push_back(nic_);

      ptr s(this);
      ret.push_back(s);
      return ret;
    }

  /*void
   simplenode::init_failure_model(double l_proc, double l_mem, double l_disk,
   double l_nic, failable::ptr ps, double p_ps_node)
   {

   SSTMAC_DEBUG << "node::init_failure_model() -- lproc = " << l_proc
   << ", lmem = " << l_mem << ", ldisk = " << l_disk << ", lnic = "
   << l_nic << "\n";

   ps->add_failable_child(compute_sched_, p_ps_node);
   compute_sched_->init_failable(l_proc, compute_sched_);

   ps->add_failable_child(mem_model_, p_ps_node);
   mem_model_->init_failable(l_mem, mem_model_);

   if (disk_)
   {
   ps->add_failable_child(disk_, p_ps_node);
   disk_->init_failable(l_disk, disk_);
   }

   ps->add_failable_child(nic_, p_ps_node);
   nic_->init_failable(l_nic, nic_);
   }*/

  }
} // end of namespace sstmac.
