/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_COMM_MPICOMM_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_COMM_MPICOMM_H_INCLUDED

#include <sstmac/software/libraries/mpi/mpi_types/mpiid.h>
#include <sstmac/software/process/taskid.h>
#include <sstmac/software/process/appid.h>
#include <sstmac/software/libraries/mpi/mpi_comm/mpicommid.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/ami/environment.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/software/process/tasklist.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpitag.h>
#include <boost/lexical_cast.hpp>
#include <vector>
#include <iosfwd>

//#include "Libraries/MPI/MPI_comm/mpicommdata.h"

namespace sstmac
{
  namespace sw
  {

    // Forward declaration of important types.
    //class mpicommdata;

    /**
     * An MPI communicator handle.
     */
    class mpicomm : public ptr_type
    {
    protected:

      friend std::ostream&
      operator<<(std::ostream &os, const mpicomm::ptr &comm);

      friend class mpicommfactory;
      /// The data (possibly shared) associated with this communicator.
      //boost::shared_ptr<const mpicommdata> data_;

      // the definitely-shared environment (softwareid <--> nodeaddress)
      environment::const_ptr env_;

      /// The tasks participating in this communicator.  This is only used for an mpicomm::ptr which is NOT WORLD_COMM.
      tasklist::const_ptr peers_;

      mutable mpitag next_collective_tag_;

      appid aid_;

      mpicomm() :
        peers_(), next_collective_tag_(10235), id_(-1), rank_(-1) //, appid_(-1)
      {
      }

      /// Hello.
      mpicomm(
          const mpicommid &id, //const appid &aid,
          const mpiid &rank, const tasklist::const_ptr peers,
          const environment::const_ptr &env, const appid &aid) :
        env_(env), peers_(peers), next_collective_tag_(10235), id_(id),
            rank_(rank), aid_(aid)
      {
      }

    public:

      typedef boost::intrusive_ptr<mpicomm> ptr;

      static ptr
      construct(
          const mpicommid &id, //const appid &aid,
          const mpiid &rank, const tasklist::const_ptr peers,
          const environment::const_ptr &env, const appid &aid)
      {
        return ptr(new mpicomm(id, rank, peers, env, aid));
      }

      static ptr
      construct() {
        return ptr(new mpicomm());
      }

      /// Goodbye.
      virtual
      ~mpicomm()
      {
      }

      /// This is the null communicator.
      static const mpicomm::ptr comm_null;

      mpicommid id_;
      mpiid rank_;

      virtual std::string
      to_string() const
      {
        return "mpicomm(" + rank().to_string() + ")";
      }

      /// The rank of this peer in the communicator.
      const mpiid&
      rank() const
      {
        return rank_;
      }

      /// The size of the communicator.
      mpiid
      size() const
      {
        if (id_.id < 0)
          {
            throw nullerror("mpicomm: trying to call size() on a null mpicomm");
          }
        if (!peers_)
          {
            throw nullerror("mpicomm: peers_ is null for some reason in size()");
          }
        return mpiid(peers_->size());
      }

      /// The identifier for this communicator.
      /// To be used to tag messages to/from this communicator.
      const mpicommid&
      id() const
      {
        return id_;
      }

      const appid&
      get_app() const
      {
        return aid_;
      }

      //
      // Get a unique tag for a collective operation.
      //
      mpitag
      next_collective_tag() const
      {
        next_collective_tag_.id++;
        return next_collective_tag_;
      }

      /// The task index of the caller.
      taskid
      my_task() const
      {
        if (id_.id < 0)
          throw illformederror("mpicomm::my_task:  NULL communicator.");
        if (rank_.id < 0 || size_t(rank_.id) >= env_->get_app_count())
          throw illformederror(
              std::string("mpicomm::my_task:  Invalid rank ")
                  + boost::lexical_cast<std::string>(rank_));

        return peers_->at(rank_.id);
      }

      /// The task index of the given peer.
      taskid
      peer_task(const mpiid &rank) const
      {
        if (id_.id < 0)
          {
            throw nullerror(
                "mpicomm: trying to call peer_task() on a null mpicomm");
          }
        if (!peers_)
          {
            throw nullerror(
                "mpicomm: peers_ is null for some reason in peer_task()");
          }
        return peers_->at(rank.id);
      }

      nodeaddress::ptr
      my_node() const
      {
        if (id_.id < 0)
          {
            throw nullerror(
                "mpicomm: trying to call my_node() on a null mpicomm");
          }
        if (!env_)
          {
            throw nullerror(
                "mpicomm: evn_ is null for some reason in my_node()");
          }

        taskid tid = my_task();
        //softwareid sid(appid_, tid);
        return env_->get_node_address(tid);
      }

      /// The list of nodes involved in this communicator.
      /// Indexing is done by mpiid::rank().id.
      nodeaddress::ptr
      node_at(const mpiid &rank) const
      {
        if (id_.id < 0)
          {
            throw nullerror(
                "mpicomm: trying to call node_at() on a null mpicomm");
          }
        if (!env_)
          {
            throw nullerror(
                "mpicomm: evn_ is null for some reason in node_at()");
          }

        taskid tid = peer_task(rank);
        //softwareid sid(appid_, tid);
        return env_->get_node_address(tid);
      }

      /// Equality comparison.
      inline bool
      operator==(const mpicomm::ptr &other) const
      {
        return ((rank_ == other->rank_) && (id_ == other->id_));
      }

      /// Inequality comparison.
      inline bool
      operator!=(const mpicomm::ptr &other) const
      {
        return !this->operator==(other);
      }

    };

    /// Allow communicators in sorted containers.
    inline bool
    operator<(const mpicomm::ptr &a, const mpicomm::ptr &b)
    {
      return (a->id() == b->id() ? a->rank() < b->rank() : a->id() < b->id());
    }

    /// Fairly self-explanatory.
    std::ostream&
    operator<<(std::ostream &os, const mpicomm::ptr &comm);

  }
} //end of namespace sstmac

#endif
