/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_COMM_MPICOMM_CART_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_COMM_MPICOMM_CART_H_INCLUDED

#include <sstmac/software/libraries/mpi/mpi_comm/mpicomm.h>

namespace sstmac
{
  namespace sw
  {

    /**
     * An MPI_Cart communicator handle.
     */
    class mpicomm_cart : public mpicomm
    {
    protected:
      int ndims_;
      std::vector<int> dims_;
      std::vector<int> periods_;
      int reorder_;

      /// Hello.
      mpicomm_cart(
          const mpicommid &id, //const appid &aid,
          const mpiid &rank, const tasklist::const_ptr peers,
          const environment::const_ptr &env, const appid &aid, int ndims,
          int *dims, int *periods, int reorder) :
        mpicomm(id, rank, peers, env, aid), ndims_(ndims), reorder_(reorder)
      {

        for (int i = 0; i < ndims; i++)
          {
            dims_.push_back(dims[i]);
            periods_.push_back(periods[i]);
          }

      }
    public:
      typedef boost::intrusive_ptr<mpicomm_cart> ptr;

      static ptr
      construct(
          const mpicommid &id, //const appid &aid,
          const mpiid &rank, const tasklist::const_ptr peers,
          const environment::const_ptr &env, const appid &aid, int ndims,
          int *dims, int *periods, int reorder)
      {
        return ptr(
            new mpicomm_cart(id, rank, peers, env, aid, ndims, dims, periods,
                reorder));
      }

      /// Goodbye.
      virtual
      ~mpicomm_cart()
      {
      }

      int
      get_dim(int i) const
      {
        if (i > dims_.size())
          return -1;
        return dims_[i];
      }

      int
      get_period(int i) const
      {
        if (i > periods_.size())
          return -1;
        return periods_[i];
      }

      void
      set_coords(int rank, int* coords)
      {
        int prev = 1;
        for (int i = 0; i < dims_.size(); i++)
          {
            int co = (rank / prev) % dims_[i];

            coords[i] = co;
            prev = prev * dims_[i];
          }
      }

      int
      get_rank(int* coords)
      {
        int rank = 0;
        int prev = 1;
        for (int i = 0; i < dims_.size(); i++)
          {
            rank += coords[i] * prev;
            prev *= dims_[i];
          }

        return rank;
      }

      int
      shift(int dir, int dis)
      {

        if (dir >= dims_.size())
          {
            std::cout << "dir: " << dir << "\n";
            throw ssterror("mpicomm_cart::shift - dir is too big");
          }
        int coords[dims_.size()];
        set_coords(rank_.id, coords);
        coords[dir] += dis;

        if (coords[dir] >= dims_[dir])
          {
            if (periods_[dir])
              {
                coords[dir] = coords[dir] % dims_[dir];
                return get_rank(coords);
              }
            else
              {
                return -1;
              }

          }
        else if (coords[dir] < 0)
          {
            if (periods_[dir])
              {
                coords[dir] = dims_[dir] - (-coords[dir]) % dims_[dir];
                return get_rank(coords);
              }
            else
              {
                return -1;
              }
          }
        else
          {
            return get_rank(coords);
          }

      }

    };
  } //end of namespace sstmac
}
#endif
