#include <sstmac/software/libraries/mpi/mpi_protocol/mpiprotocol.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue-sendrequest.h>
#include <sstmac/software/libraries/mpi/mpiserver.h>

namespace sstmac {
    namespace sw {


void
eager1_rdma::send_header(const mpiqueue::ptr &queue, const mpimessage::ptr &msg)
{
    SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
          << "]:: send_eager1_rdma_header done \n";

    msg->set_needs_ack(false);
    /** This begins by copying into a tmp buffer */
    if (queue->server()->get_os()->get_current_threadid() == -1) //this is from the mpi server
    {
        queue->server()->buffered_send(msg);
    }
    else
    {
        /** we have a negative count for the message */
        long bytes = -msg->count() * msg->type().size;
        queue->user_lib_mem()->copy(bytes);
        queue->server()->send(msg);
    }

    /** the send will have copied into a temp buffer so we can 'ack' the buffer for now */
    mpiqueue::send_needs_ack_t::iterator it, end = queue->send_needs_nic_ack_.end();
    for (it = queue->send_needs_nic_ack_.begin(); it != end; ++it)
    {
         //horrible hack for now
        if ((*it)->matches(msg))
        {
            SSTMAC_DEBUG << "eager1 nic ack match found \n";
            // Match.
            mpiqueue::sendrequest::ptr sreq = *it;
            sreq->complete(msg);
            /** we can complete the send request and allow
                the program to continue, but we have to wait
                for buffer resources to be completed by the rdma get */
            sreq->wait_for_buffer();
            return;
        }
    }
    sst_throw(illformederror, "eager1 protocol could not find request to complete");
}

void
eager1_rdma::handle_mpi_ack(
    const mpiqueue::ptr &queue,
    const mpimessage::ptr &msg,
    const mpiqueue::sendrequest::ptr &req
)
{
    if (msg->ssend())
        req->complete(msg);
    else
        sst_throw(illformederror, "eager message should not be receiving mpi ack");
}

void
eager1_rdma::handle_header(const mpiqueue::ptr &queue, const mpimessage::ptr &msg)
{
    mpiqueue::recvrequest::ptr req = queue->find_pending_request(msg);
    if (req)
      { //we can post an RDMA get request direct to the buffer
        SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
              << "]:: eager1 got lucky and found matching request\n";
        //make sure to put the request back in, but alert it
        //that it should expect a data payload next time
        req->set_category(mpimessage::data);
        queue->pending_message_.push_front(req);
      }
    else
      {
        SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
              << "]:: eager1 is unlucky and did not find matching request\n";
        msg->set_protocol(mpiprotocol::eager1_rdma_doublecpy_protocol);
      }
    queue->notify_probes(msg);

    /** we need to send an RDMA GET request back to the original node */
    msg->mutate(sst_message::GNI_HEADER_TO_RDMA_GET_REQ);
    msg->resequence(mpimessage::rdma_ignore_sequence);

    queue->server()->send(msg);

    SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
          << "]:: handle_eager1_rdma_header done \n";
}

void
eager1_rdma::recv_header(
    const mpiqueue::ptr &queue,
    const mpimessage::ptr &msg,
    const mpiqueue::recvrequest::ptr& req
)
{
    sst_throw(illformederror, "eager1_rdma protocol should not have recv-request handle header");
}

void
eager1_rdma::recv_payload(
    const mpiqueue::ptr &queue,
    const mpimessage::ptr &msg,
    const mpiqueue::recvrequest::ptr& req
)
{
}

void
eager1_rdma_singlecpy::handle_payload(const mpiqueue::ptr &queue, const mpimessage::ptr &msg)
{
    SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
          << "]: incoming eager1 singlecpy payload\n";
    mpiqueue::recvrequest::ptr req = queue->find_pending_request(msg);
    if (!req)
      throw illformederror(
          "running singlecpy eager1 protocol, but no matching request found");
    //We have RDMA GOT direct into the buffer - just complete
    req->handle(msg);
}

void
eager1_rdma_doublecpy::handle_payload(const mpiqueue::ptr &queue, const mpimessage::ptr &msg)
{
    SSTMAC_DEBUG << "mpiqueue[" << queue->id_string()
          << "]: incoming eager1 doublecpy payload\n";
    //the recv request is expecting a user message - update the message to match
    msg->cat(mpimessage::user);
    mpiqueue::recvrequest::ptr req = queue->find_pending_request(msg);
    if (req) //just finish things off by copying buffers
      {
        queue->server()->buffered_recv(msg ,req);
      }
    else
      {
        //the recv hasn't been posted yet
        queue->need_recv_.push_back(msg);
      }
}


bool
eager1_rdma::handshake_only() const
{
    //initial send is handshake only
    return true;
}

bool
eager1_rdma::need_mpi_ack() const
{
    return false;
}
    }
}
