/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_protocol/mpiprotocol.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue-sendrequest.h>
#include <sstmac/software/libraries/mpi/mpiserver.h>
#include <sstmac/software/libraries/mpi/mpimessage.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>

namespace sstmac { namespace sw
{

  // Pointer types.
  
  

  //
  // Hello.
  //
  mpiqueue::sendrequest::sendrequest(const mpirequest_t key, std::string libn,
      const mpiqueue::ptr& queue, int64_t envelope,
      int64_t mintrans, int count, const mpitype &type, const mpiid &dest,
      const mpitag& tag, const mpicomm::ptr &comm, bool rsend, bool ssend,
      mpimessage::category cat, int seqno, const messageid &id,
      const eventhandler::ptr& completion,
      const payload::const_ptr& content) :
      queue_(queue), envelope_(envelope),
        mintrans_(mintrans), count_(count), type_(type), dest_(dest),
        tag_(tag), comm_(comm), rsend_(rsend), ssend_(ssend), cat_(cat),
        seqno_(seqno), id_(id), completion_(completion), content_(content),
        libname_(libn), key_(key)
  {
  }

  //
  // Hi there.
  //
  mpiqueue::sendrequest::ptr
  mpiqueue::sendrequest::construct(const mpirequest_t key, std::string libn,
      const mpiqueue::ptr& queue, int64_t envelope,
      int64_t mintrans, int count, const mpitype &type,
      const mpiid &dest, const mpitag& tag, const mpicomm::ptr &comm, bool rsend,
      bool ssend, mpimessage::category message_type, int seqno,
      const messageid &id, const eventhandler::ptr& completion,
      const payload::const_ptr& content)
  {

    sendrequest::ptr retval(new sendrequest(key, libn, queue,
        envelope, mintrans, count, type, dest, tag, comm, rsend, ssend,
        message_type, seqno, id, completion, content));

    return retval;
  }

  //
  // Goodbye.
  //
  mpiqueue::sendrequest::~sendrequest() throw ()
  {
   // std::cout << "calling sendrequest destructor" << std::endl;
  }

  //
  // Go.
  //
  void
  mpiqueue::sendrequest::start(const mpiserver::ptr& lib)
  {

    if (lib == 0)
        sst_throw(nullerror, "mpiqueue::sendrequest::start:  null mpiserver pointer");

    mpimessage::ptr mess;
    timestamp sendtime(1E-9); // in the future, can make this separate send to the ami or something
    const taskid pt = comm_->peer_task(dest_);
    nodeaddress::ptr src = comm_->node_at(comm_->rank());
    nodeaddress::ptr dst = comm_->node_at(dest_);
    bool intranode = src->equals(dst);
    long bytes = count_ * int64_t(type_.size);
    mpiprotocol::ptr protocol = queue_->get_protocol(bytes, intranode, rsend_);
    bool handshake_only = protocol->handshake_only();
    int msg_count = handshake_only ? -count_ : count_; //negative size indicates handshake
    mess = mpimessage::construct(libname_, sendtime, envelope_, mintrans_,
        msg_count, type_, comm_->rank(), dest_, tag_, comm_->id(), seqno_, id_,
        ssend_, cat_, comm_->my_task(), pt, comm_->get_app(), content_, protocol);
    mess->fromaddr_ = src;
    mess->toaddr_ = dst;


    if (ssend_ || protocol->need_mpi_ack()){
        queue_->request_ack(mess, sendrequest::ptr(this));
    }
    protocol->send_header(queue_, mess);

    if (!lib->nic_send_ack_supported())
    {
        complete(mess);
    }
  }

  //
  // We just received an ack.
  //
  void
  mpiqueue::sendrequest::ack(const mpimessage::ptr& response)
  {
      if (response->seqnum() != seqno_)
      {
          throw illformederror("mpiqueue::sendrequest::ack:  Received a "
            "non-matching ack.");
      }
      response->get_protocol()->handle_mpi_ack(queue_, response, sendrequest::ptr(this));
      if (ssend_)
          completion_->handle(response);
  }

  //
  // Test whether we match a given message, for nic send acks.
  //
  bool
  mpiqueue::sendrequest::matches(const mpimessage::ptr& message) const
  {
    bool sameid = (seqno_ == message->seqnum());
    return (sameid && comm_->rank().matches(message->source())
            && tag_.matches(message->tag()));
  }

  void
  mpiqueue::sendrequest::wait_for_buffer()
  {
      /**
        CHANGE! At some point, we may want to track buffer
        resources for MPI. For now, let's not worry about buffers.
        Just clear the completion handlers and use this as a
        placeholder for clearing nic acks
      */
    key_ = 0;
    completion_ = 0;
  }

  void
  mpiqueue::sendrequest::complete(const mpimessage::ptr &msg)
  {
    if (key_)
        key_->complete(msg);

    if (completion_)
      completion_->handle(msg);

  }

} } // end of namespace sstmac.
