/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue-recvrequest.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue-sendrequest.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue-proberequest.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/libraries/mpi/mpiserver.h>
#include <sstmac/software/process/appid.h>
#include <sstmac/software/libraries/mpi/mpi_implementation/mpi_implementation.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/software/libraries/mpi/mpi_protocol/mpiprotocol.h>

#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>
#include <stdint.h>

namespace sstmac
{
  namespace sw
  {

    // Pointer types.
    using namespace std;

    debug mpiqueue::dbg_;

    stat_spyplot::ptr mpiqueue::spy_;

    stat_spyplot::ptr mpiqueue::spy_bytes_;

    bool
    mpiqueue::sortbyseqnum::operator()(const mpimessage::ptr &a,
        const mpimessage::ptr &b) const
    {
      return (a->seqnum() < b->seqnum());
    }

    void
    mpiqueue::set_eventmanager(const eventmanager::ptr& m)
    {
      if (!spy_)
        {
          spy_ = stat_spyplot::construct("<stats> spyplot", "mpi_spyplot.csv");
          m->register_stat(spy_);

          spy_bytes_ = stat_spyplot::construct("<stats> spyplot",
              "mpi_spyplot_bytes.csv");
          m->register_stat(spy_bytes_);
        }
      server_->set_eventmanager(m);
    }

    //
    // Hi there.
    //
    mpiqueue::mpiqueue(const mpiserver::ptr& server, taskid myid,
        int64_t mintrans) :
      server_(server), mintrans_(mintrans), myid_(myid), next_id_(0),
          mpi_impl_(0)
    {
      mpiprotocol::init();

      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> mpi | mpiqueue", logger::magenta);
          logger::register_logger(&dbg_);
        }

      if (server == 0)
        {
          throw nullerror("mpiqueue(...):  null server pointer.");
        }

      // need to debug dangling messages.
      next_id_ = messageid(myid_.id * int64_t(4000000000LL));

      std::string libname = "mpiqueue-user-lib-mem" + myid.to_string();
      user_lib_mem_ = lib_compute_memmove::construct(libname);
      server_->get_os()->register_lib(user_lib_mem_);

      std::string mpitype = sstmac_env::params->get_optional_param(
          "mpi_implementation", "basic");
      mpi_impl_ = mpi_implementation::construct(mpitype);
    }

    void
    mpiqueue::unregister(const softwareid& sid){
      server_->get_os()->unregister_lib(user_lib_mem_);
      user_lib_mem_->unregister_all_libs();
      //server_->unregister(sid);
    }

    //
    // Hello.
    //
    mpiqueue::ptr
    mpiqueue::construct(const mpiserver::ptr& server, taskid myid,
        int64_t mintrans)
    {
      mpiqueue::ptr retval(new mpiqueue(server, myid, mintrans));
      return retval;
    }

    //
    // Goodbye.
    //
    mpiqueue::~mpiqueue() throw ()
    {
    }

    //
    // Send data.
    //
    void
    mpiqueue::send(const mpirequest_t key, int count, const mpitype &type,
        const mpiid &dest, const mpitag &tag, const mpicomm::ptr &comm,
        bool rsend, bool ssend, mpimessage::category message_type,
        const eventhandler::ptr &completion, const payload::const_ptr& content)
    {
        SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::recv("
              << "count=" << count
              << ", type=" << type.to_string()
              << ", dest=" << dest.to_string()
              << ", tag=" << tag.to_string()
              << ", comm=" << comm->to_string()
              << ", kind=" << (int) message_type
              << ")\n";

      taskid tid = comm->peer_task(dest);

      sendrequest::ptr req = sendrequest::construct(key,
          server_->get_lib_name(), mpiqueue::ptr(this),
          mpi_impl_->get_envelope_size(), mintrans_, count, type, dest, tag,
          comm, rsend, ssend, message_type, next_outbound_[tid]++, next_id_++,
          completion, content);

      if (server_ == 0)
        {
          throw illformederror("mpiqueue::send:  null server pointer.");
        }

      /** push this on first. important! */
      send_needs_nic_ack_.push_back(req);
      req->start(server_);

      spy_->add_one(myid_.id, dest.id);
      spy_bytes_->add(myid_.id, dest.id, count * type.size);
    }

    mpiprotocol::ptr
    mpiqueue::get_protocol(long bytes, bool intranode, bool rsend) const
    {
      return mpi_impl_->get_protocol(bytes, intranode, rsend);
    }

    //
    // Receive data.
    //
    void
    mpiqueue::recv(const mpirequest_t key, int count, const mpitype &type,
        const mpiid &source, const mpitag &tag, const mpicomm::ptr &comm,
        mpimessage::category message_type, eventhandler::ptr &completion)
    {
        SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::recv("
            << "count=" << count
            << ", type=" << type.to_string()
            << ", source=" << source.to_string()
            << ", tag=" << tag.to_string()
            << ", comm=" << comm->to_string()
            << ", kind=" << (int) message_type
            << ")\n";

      recvrequest::ptr req = recvrequest::construct(key, mpiqueue::ptr(this),
          count, type, source, tag, comm, message_type, completion);
      // Figure out whether we already have a matching mesage.

      need_recv_t::iterator it, end = need_recv_.end();
      for (it = need_recv_.begin(); it != end; ++it)
        {
          mpimessage::ptr mess = *it;
          if (req->matches(mess))
            {
              SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                    << "]::recv:  Matching message found.\n";
              // We're good to go.

              need_recv_.erase(it);
              if (mess->count() > 0) //we need a memcpy from a buffer
                {
                  long dst_thread = server_->get_os()->get_current_threadid();
                  if (dst_thread == thread::main_thread)
                    {
                      /** the server needs to perform the memcpy */
                      server_->buffered_recv(mess, req);
                    }
                  else
                    {
                      /** Again, just guess for now that the message is still in cache.
                          However, the memory has been copied wherever the mpi buffers
                          are set up which is almost certainly on die 0.  This is
                          configured by just declaring 'nic thread' to let the memory
                          know to look for the buffer wherever the nic put it
                      */
                      user_lib_mem_->write(mess->get_byte_length(), thread::nic_thread, dst_thread);
                      req->handle(mess);
                    }
                }
              else
                {
                  //just handle the request - this is a handshake
                  req->handle(mess);
                }
              return;
            }
        }

      // If we get here, we still need to wait for the message.
      SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
            << "]::recv:  Matching message not found\n";
      pending_message_.push_back(req);
    }

    //
    // Ask for a notification when a message with the given signature arrives.
    //
    void
    mpiqueue::probe(const mpirequest_t key, const mpicomm::ptr &comm,
        const mpiid &source, const mpitag &tag, eventhandler::ptr &completion)
    {
      SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::probe("
            << comm->to_string() << ", " << source.to_string() << ", "
            << tag.to_string() << ", " << ")\n";

      proberequest::ptr req = proberequest::construct(key, comm, source, tag,
          completion);
      // Figure out whether we already have a matching message.
      need_recv_t::iterator it, end = need_recv_.end();
      for (it = need_recv_.begin(); it != end; ++it)
        {
          if ((*it)->cat() == mpimessage::user && req->matches(*it))
            {
              // We're good to go.
              req->complete(*it);
              return;
            }
        }
      // If we get here, we still need to wait for the message.
      probelist_.push_back(req);
    }

    //
    // Immediate-mode probe for a message with the given signature.
    //
    mpistatus::ptr
    mpiqueue::iprobe(const mpicomm::ptr &comm, const mpiid &source,
        const mpitag &tag)
    {
      SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::iprobe("
            << comm->to_string() << ", " << source.to_string() << ", "
            << tag.to_string() << ")\n";
      proberequest::ptr req = proberequest::construct(mpirequest_t(), comm,
          source, tag, eventhandler::ptr());
      need_recv_t::iterator it, end = need_recv_.end();
      for (it = need_recv_.begin(); it != end; ++it)
        {
          if ((*it)->cat() == mpimessage::user && req->matches(*it))
            {
              // This is it.
              return (*it)->build_status();
            }
        }
      // If we got here, there was no match
      return mpistatus::ptr();
    }

    //
    // Used when a receive operation sends an ack.
    //
    void
    mpiqueue::send_mpi_ack(const mpimessage::ptr& message,
        const recvrequest::ptr& operation)
    {
      SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::send_ack("
            << message << ", " << operation << ")\n";

      message->header_to_ack();
      // OK,
      if (operation) // If this is non-null, the sender expects a response to its ack.
          mpi_ack_sent_[message->id()] = operation;

      server_->send(message);
    }

    //
    // Used when a send operation requests an ack for the given message.
    //
    void
    mpiqueue::request_ack(const mpimessage::ptr& message,
        const sendrequest::ptr& operation)
    {
      SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]::request_ack("
            << message->to_string() << ", " << ")\n" << "  :  Message id: "
            << message->id().to_string() << "\n";

      mpi_ack_needed_[message->id()] = operation;
    }

    //
    // CALLBACK used by mpiserver when this object has a message.
    //
    void
    mpiqueue::incoming_message(const mpimessage::ptr& message)
    {
      SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
            << "]::incoming_message(" << ", " << message->to_string() << ")\n";

      if (message->get_type() == sst_message::SEND_ACK)
        {
          this->handle_nic_ack(message);
          return;
        }

      /// These are the types of messages we may expect:
      /// (1)  Response to handshake request (we have the requestor).
      /// (2)  Data delivery following a handshake ack (we have the receiver)
      /// (3)  Acks from the NIC that a message has left
      /// (4)  New messages (either eager-send or a handshake request).
      /// Only category (4) is subject to order requirements.  The other
      /// categories get handled immediately

      switch (message->cat())
        {
      case mpimessage::ack:
        this->incoming_rendezvous_ack(message);
        break;
      case mpimessage::data:
        message->get_protocol()->handle_payload(mpiqueue::ptr(this), message);
        break;
      case mpimessage::user:
      case mpimessage::collective:
      case mpimessage::internal:
        this->incoming_new_message(message);
        break;
        }
    }

    //
    // Handle a response to an ack posted from this queue.
    //
    void
    mpiqueue::incoming_rendezvous_ack(const mpimessage::ptr &message)
    {
      const messageid &id = message->id();
      SSTMAC_DEBUG << "  :  Incoming ack from task "
            << message->source_task().to_string() << " to task "
            << message->dest_task().to_string() << " with id "
            << id.to_string() << "\n" << " : Finding ack for "
            << message->id().to_string() << "\n";

      ack_needed_t::iterator it = mpi_ack_needed_.find(message->id());
      if (it != mpi_ack_needed_.end())
        {
          // We have a winner.
          SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                << "]::incoming_message:  " << "Ack with matching request.\n";
          sendrequest::ptr req = it->second;
          mpi_ack_needed_.erase(it);
          req->ack(message);
        }
      else
        {
          // No match for this ack (this should not happen).
          SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                << "]::incoming_message:  ERROR:  " << "Ack with no match.\n";
          throw illformederror("pmiqueue::incoming_message: Ack with no match.");
        }
    }

    //
    // Handle a new incoming message.  Can be either an eager send (with data)
    // or a new handshake request
    //
    void
    mpiqueue::incoming_new_message(const mpimessage::ptr &message)
    {
      SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
            << "]: incoming_new_message \n";
      taskid tid = message->source_task();
      if (message->seqnum() == next_inbound_[tid])
        {
          SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                << "]: seqnum matched expected " << next_inbound_[tid]
                << " for task " << tid.id << "\n";
          // Handle this message.
          handle_new_message(message);
          ++next_inbound_[tid];
          SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                << "]: advanced to seqnum " << next_inbound_[tid]
                << " for task " << tid.id << "\n";
          // Handle any messages that have been freed by the arrival of this one
          if (!held_[tid].empty())
            {
              SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                    << "]: there are old messages for task " << tid.id << "\n";

              hold_list_t::iterator it = held_[tid].begin(), end =
                  held_[tid].end();
              while (it != end)
                {
                  mpimessage::ptr mess = *it;
                  if (mess->seqnum() <= next_inbound_[tid])
                    {
                      //it = held_[tid].erase(it);
                      it++;
                      handle_new_message(mess);
                      ++next_inbound_[tid];

                    }
                  else
                    {
                      break;
                    }

                }
              // Now erase all the completed messages from the held queue.
              held_[tid].erase(held_[tid].begin(), it);
            }
        }
      else if (message->seqnum() == mpimessage::rdma_ignore_sequence)
      {
          SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                << "]: got rdma seqnum\n";
          handle_new_message(message);
      }
      else
        {
          SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                       << "]: putting message in held, seqnum " << message->seqnum()
                       << " didn't match expected " << next_inbound_[tid]
                       << " for task " << tid.id << "\n";
          held_[tid].insert(message);
        }

      SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
            << "]: incoming_new_message done \n";
    }

    void
    mpiqueue::notify_probes(const mpimessage::ptr& message)
    {
      // And, finally, mpimessage::user messages are tested for outstanding probes
      // list<proberequest::ptr> proberemove;
      SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string() << "]:probelist size: "
            << probelist_.size() << "\n";
      probelist_t::iterator pit = probelist_.begin(), pend = probelist_.end();
      while (pit != pend)
        {
          SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                << "]: testing probe \n";
          if ((*pit)->matches(message))
            {
              SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                    << "]: probe match \n";
              // Match.
              proberequest::ptr preq = *pit;

              pit = probelist_.erase(pit);
              preq->complete(message);
            }
          else
            {
              pit++;
            }

          pend = probelist_.end();
        }
    }

    void
    mpiqueue::handle_new_message(const mpimessage::ptr &message)
    {
        message->get_protocol()->handle_header(mpiqueue::ptr(this), message);
    }

    //
    // Complete an inbound message.
    //
    mpiqueue::recvrequest::ptr
    mpiqueue::find_pending_request(const mpimessage::ptr &message)
    {
      std::list<recvrequest::ptr> to_erase;
      recvrequest::ptr ret;
      pending_message_t::iterator it, end = pending_message_.end();
      for (it = pending_message_.begin(); it != end; ++it)
        {
          if ((*it)->is_cancelled())
            {
              to_erase.push_back(*it);
            }
          else if ((*it)->matches(message))
            {
              SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
                    << "]::handle_new_message:  Receiver is ready.\n";
              ret = *it;
              to_erase.push_back(*it);
              break;
            }
        }

      std::list<recvrequest::ptr>::iterator it2, end2 = to_erase.end();
      for(it2 = to_erase.begin(); it2 != end2; it2++){
          end = pending_message_.end();
          for(it = pending_message_.begin(); it != end; it++){
              if(*it == *it2){
                  pending_message_.erase(it);
                  break;
              }
          }
      }
      if (ret)
        return ret;

      // We get here if no match was found.
      // Messages that don't have a respondent are added to the list
      SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
            << "]::handle_new_message:  receiver not ready for "
            << message->to_string() << "\n";
      need_recv_.push_back(message);

      return recvrequest::ptr();
    }

    void
    mpiqueue::complete_nic_ack(const mpimessage::ptr& message)
    {
        send_needs_ack_t::iterator it, end = send_needs_nic_ack_.end();

        for (it = send_needs_nic_ack_.begin(); it != end; ++it)
          {
            if ((*it)->matches(message))
              {
                SSTMAC_DEBUG << "match found \n";
                // Match.
                sendrequest::ptr sreq = *it;
                send_needs_nic_ack_.erase(it);
                sreq->complete(message);
                return;
              }
          }
        sst_throw(valueerror, "no match found for nic ack on %s",
                  message->to_string().c_str());
    }

    void
    mpiqueue::handle_nic_ack(const mpimessage::ptr &message)
    {
      SSTMAC_DEBUG << "mpiqueue[" << myid_.to_string()
            << "]::handle_nic_ack...";

      /** ignore the acks from handshake messages -
          we still have work to do to send */
      if (message->count() >= 0)
        complete_nic_ack(message);
    }

    const mpiserver::ptr&
    mpiqueue::server() const
    {
        return server_;
    }

    const lib_compute_memmove::ptr&
    mpiqueue::user_lib_mem() const
    {
        return user_lib_mem_;
    }

    std::string
    mpiqueue::id_string() const
    {
        return myid_.to_string();
    }

    mpi_implementation::ptr
    mpiqueue::implementation() const
    {
        return mpi_impl_;
    }

  }
} // end of namespace sstmac
