/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpimessage.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/libraries/mpi/mpi_protocol/mpiprotocol.h>

#include <sstmac/software/process/operatingsystem.h>

#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>
#include <vector>
#include <stdlib.h>

namespace sstmac
{
  namespace sw
  {

    // Pointer types.

    //
    // Hello.
    //
    mpimessage::mpimessage(std::string libn, timestamp time, int64_t envelope,
        int64_t mintrans, int count, const mpitype &type, const mpiid &source,
        const mpiid &dest, const mpitag &tag, const mpicommid &commid,
        int seqnum, const messageid &msgid, bool ssend, int cat,
        const taskid &source_task, const taskid &dest_task, const appid &aid,
        const payload::const_ptr& content, const mpiprotocol::ptr& protocol, MESSAGE_TYPES ty) :
        sst_message(libn, time, ty, cat), envelope_(envelope), mintrans_(
            mintrans), count_(count), type_(type), source_(source), dest_(dest), tag_(
            tag), commid_(commid), seqnum_(seqnum), msgid_(msgid), ssend_(
            ssend), source_task_(source_task), aid_(aid), dest_task_(dest_task), content_(
            content), fake(false), protocol_(protocol)
    {
      if (envelope_ < 0)
        envelope_ = 0;
      if (mintrans_ < 0)
        mintrans_ = 0;

    }

    mpimessage::mpimessage() :
        sst_message(MPI), commid_(-1), fake(false), protocol_(mpiprotocol::eager0_socket_protocol)
    {
    }

    mpimessage::mpimessage(const payload::const_ptr& content) :
        sst_message(MPI), commid_(-1), content_(content), fake(true), protocol_(mpiprotocol::eager0_socket_protocol)
    {
    }

    //
    // Construct a new message.
    //
    mpimessage::ptr
    mpimessage::construct(std::string libn, timestamp time, int64_t envelope,
        int64_t mintrans, int count, const mpitype &type, const mpiid &source,
        const mpiid &dest, const mpitag &tag, const mpicommid &commid,
        int seqnum, const messageid &msgid, bool ssend, int cat,
        const taskid &source_task, const taskid &dest_task, const appid &aid,
        const payload::const_ptr& content, const mpiprotocol::ptr& protocol, MESSAGE_TYPES ty)
    {

      mpimessage::ptr retval(
          new mpimessage(libn, time, envelope, mintrans, count, type, source,
              dest, tag, commid, seqnum, msgid, ssend, cat, source_task,
              dest_task, aid, content, protocol, ty));
      return retval;
    }

    sst_message::ptr
    mpimessage::clone() const
    {
        return deep_clone();
    }

    mpimessage::ptr
    mpimessage::deep_clone() const
    {
        mpimessage::ptr ret = mpimessage::construct(lib_name_, time_, envelope_,
            mintrans_, count_, type_, source_, dest_, tag_, commid_, seqnum_,
            msgid_, ssend_, kind_, source_task_, dest_task_, aid_, content_, protocol_);

        ret->toaddr_ = toaddr_;
        ret->fromaddr_ = fromaddr_;

        return ret;
    }

    sst_message::ptr
    mpimessage::clone(MESSAGE_TYPES ty) const
    {
      mpimessage::ptr ret = mpimessage::construct(lib_name_, time_, envelope_,
          mintrans_, count_, type_, source_, dest_, tag_, commid_, seqnum_,
          msgid_, ssend_, kind_, source_task_, dest_task_, aid_, content_, protocol_, ty);

      ret->toaddr_ = toaddr_;
      ret->fromaddr_ = fromaddr_;

      return ret;
    }

    //
    // Goodbye.
    //
    mpimessage::~mpimessage() throw ()
    {
    }

    //
    // Create a status object.
    //
    mpistatus::ptr
    mpimessage::build_status() const
    {
      mpistatus::ptr stat = mpistatus::construct();
      if (!fake)
        {
          stat->set_commid(this->commid());
          stat->set_source(this->source());
          stat->set_dest(this->dest());
          stat->set_tag(this->tag());
          stat->set_send_type(this->type());
          stat->set_count(this->count_);
        }
      stat->set_content(this->content());
      return stat;
    }

    std::string
    mpimessage::to_string() const
    {
      std::stringstream ss;
      ss << "mpimessage(env=" << envelope_
          << ", mintrans=" << mintrans_
          << ", count=" << count_
          << ", type=" << type_.to_string()
          << ", src=" << source_.to_string()
          << ", dst=" << dest_.to_string()
          << ", tag=" << tag_.to_string()
          << ", commid" << commid_.to_string()
          << ", seq=" << seqnum_
          << ", ssend=" << ssend_
          << ", kind=" << (int) kind_
          << ", srctsk=" << source_task_.to_string()
          << ", dsttsk=" << dest_task_.to_string()
          << ", protocol=" << protocol_->to_string()
          << ", load=" << ((content_ == 0) ? "null" : content_->to_string())
          << ")";
      return ss.str();
    }

    void
    mpimessage::serialize(sst_serializer* ser)
    {
      //std::cout << "in mpimessage serialize" << std::endl;
      sst_message::serialize(ser);

      ser->pack((long long int) envelope_);
      ser->pack((long long int) mintrans_);
      ser->pack(count_);
      ser->pack(type_.id);
      ser->pack(source_.id);
      ser->pack(dest_.id);
      ser->pack(tag_.id);
      ser->pack(commid_.id);
      ser->pack(seqnum_);
      ser->pack(msgid_.id);
      ser->pack(ssend_);
      ser->pack(kind_);
      ser->pack(source_task_.id);
      ser->pack(dest_task_.id);

      bool hasContent = false;
      if (content_)
        {
          hasContent = true;
          ser->pack(hasContent);
          // std::cout << "serializing content: " << content_->to_string() << std::endl;
          content_->serialize(ser);
        }
      else
        {
          // std::cout << "no content to serialize " << std::endl;
          ser->pack(hasContent);
        }

      // std::cout << "completed mpimessage serialize" << std::endl;
    }
    sst_message::ptr
    mpimessage::deserialize(sst_serializer* ser)
    {
      // std::cout << "in mpimessage deserialize" << std::endl;

      mpimessage::ptr ret = mpimessage::construct();
      ret->do_deser(ser);
      long long int temp;
      ser->unpack(temp);
      ret->envelope_ = temp;
      ser->unpack(temp);
      ret->mintrans_ = temp;
      ser->unpack(ret->count_);
      ser->unpack(ret->type_.id);
      ser->unpack(ret->source_.id);
      ser->unpack(ret->dest_.id);
      ser->unpack(ret->tag_.id);
      ser->unpack(ret->commid_.id);
      ser->unpack(ret->seqnum_);
      ser->unpack(ret->msgid_.id);
      ser->unpack(ret->ssend_);
      ser->unpack(ret->kind_);
      ser->unpack(ret->source_task_.id);
      ser->unpack(ret->dest_task_.id);

      bool hasContent;
      ser->unpack(hasContent);
      if (hasContent)
        {

          ret->content_ = sst_serializer::deserialize_content(ser);
        }
      else
        {
          // std::cout << "no content to deserialize" << std::endl;
          ret->content_ = payload::null();
        }

      // std::cout << "completed mpimessage deserialize" << std::endl;

      return ret;
    }

    int64_t
    mpimessage::bytes() const
    {
        // Pretend that envelope info is 12 bytes.
        // A handshake request is, therefore, only 12 bytes.
        if (count_ > 0)
        {
            int64_t sbytes = envelope_ + count_ * int64_t(type_.size);
            return (sbytes > mintrans_ ? sbytes : mintrans_);
        }
        else
            return (envelope_ > mintrans_ ? envelope_ : mintrans_);
    }

    void
    mpimessage::mutate(mutation_t mut)
    {
        if (mut == sst_message::RDMA_GET_REQ_TO_RSP)
        {
            if (count_ > 0)
                throw illformederror("mpimessage: invalid rdma request generation from real payload");
            count_ = -count_;
            msgtype_ = RDMA_GET_RSP;
            kind_ = mpimessage::data;
            reverse();
        }
        else if (mut == sst_message::GNI_HEADER_TO_RDMA_GET_REQ)
        {
            msgtype_ = RDMA_GET_REQ;
            reverse();
        }
        else
        {
            sst_throw(valueerror, "invalid mutation type received by mpimessage: %s", sst_message::tostr(mut));
        }
    }

    void
    mpimessage::reverse()
    {
        //also flip the addresses
        nodeaddress::ptr dst = fromaddr_;
        nodeaddress::ptr src = toaddr_;
        toaddr_ = dst;
        fromaddr_ = src;

        taskid srcid = source_task_;
        taskid dstid = dest_task_;
        source_task_ = dstid;
        dest_task_ = srcid;
    }

    void
    mpimessage::header_to_ack()
    {
        reverse();
        kind_ = mpimessage::ack;
        set_needs_ack(false);
    }

    void
    mpimessage::ack_to_payload()
    {
        reverse();
        kind_ = mpimessage::data;
        set_needs_ack(true);
        count_ = -count_;
    }

    mpiprotocol::ptr
    mpimessage::get_protocol() const
    {
        return protocol_;
    }

    void
    mpimessage::set_protocol(const mpiprotocol::ptr& protocol)
    {
        protocol_ = protocol;
    }

    long
    mpimessage::source_thread(const operatingsystem::ptr& os) const
    {
        return os->get_task_threadid(source_task_);
    }

    long
    mpimessage::dest_thread(const operatingsystem::ptr& os) const
    {
        return os->get_task_threadid(dest_task_);
    }

    void
    mpimessage::resequence(int seqnum)
    {
        seqnum_ = seqnum;
    }

  }
} // end of namespace sstmac
