#include <sstmac/software/process/graphviz.h>
#include <sstmac/common/basicstringtokenizer.h>
#include <sstmac/common/config.h>
#include <sstmac/software/process/operatingsystem.h>

#if HAVE_EXECINFO_H
#include <execinfo.h>
#endif

#if HAVE_DLFCN_H
#include <dlfcn.h>
#endif

#include <sstream>
#include <vector>

#define heisenbug std::cout << __LINE__ << " " << __FILE__ << std::endl

namespace sstmac {
    namespace sw {

#define BACKTRACE_NFXN 50
typedef void* graphviz_trace[BACKTRACE_NFXN];
std::list<void**> trace_pool_;


char debug[2048];
char cmd[512];

void
GraphViz::Trace::add_call(const std::string& fxn, int ncalls, long count)
{
    graphviz_call& call = calls_[fxn];
    call.first += ncalls;
    call.second += count;
}

void
GraphViz::Trace::add_self(long count)
{
    self_ += count;
}

std::string
GraphViz::Trace::summary() const
{
    std::stringstream sstr;
    sstr << "fn=" << fxn_ << "\n";
    sstr << 0 << " " << self_ << "\n";
    std::map<std::string, graphviz_call>::const_iterator it = calls_.begin(), end = calls_.end();
    for ( ; it != end; ++it)
    {
        const std::string& fxn = it->first;
        const graphviz_call& call = it->second;
        sstr << "cfn=" << fxn << "\n";
        sstr << "calls=" << call.first << " " << 0 << "\n";
        sstr << 0 << " " << call.second << "\n";
    }
    return sstr.str();
}

std::string
GraphViz::Trace::fxn() const
{
    return fxn_;
}


GraphViz::Trace::ptr
GraphViz::Trace::construct(const GraphViz::ptr& parent, const std::string& fxn)
{
    return new Trace(parent, fxn);
}

#if HAVE_GRAPHVIZ
void
GraphViz::get_info(
            void* addr,
            std::string& fxn)
{
    fxn = fxns_[addr];
    if (fxn.size() != 0)
        return;

#if HAVE_DLFCN_H
    Dl_info info;
    int err = dladdr(addr, &info);
    if (err == 0){
        fprintf(stderr, "failed on %p\n", addr);
        abort();
    }

    FILE *fp;

#if HAVE_ATOS
    snprintf(cmd, sizeof(cmd), "atos -o %s -l %p %p", info.dli_fname, info.dli_fbase, addr);
    fp = popen(cmd, "r");
    if (fp) {
        fgets(debug, sizeof(debug), fp);
    }
    pclose(fp);

    std::string output = debug;
    std::deque<std::string> tok;
    std::string splitter = "(";
    pst::BasicStringTokenizer::tokenize(output, tok, splitter);
    std::deque<std::string>::const_iterator it, end = tok.end();
    int count = 0;
    for (it = tok.begin(); it != end; ++it, ++count)
    {
        if (count == 0)
        {
            fxn = *it;
        }
        /**
        else if (count == 3)
        {
            std::string colon = ":";
            std::deque<std::string> subtok;
            pst::BasicStringTokenizer::tokenize(*it, subtok, colon);
            std::deque<std::string>::const_iterator subit = subtok.begin();
            filename_ = *subit;
            ++subit;
            std::string line_tmp = *subit;
            line_ = atoi(line_tmp.substr(0, line_tmp.size()).c_str());
        }
        */
    }
#elif HAVE_ADDR_TO_LINE
    throw illformederror("Cannot use GraphViz utility with addr2line yet.  Not yet implemented.");
#else
    throw illformederror("Cannot use GraphViz utility. Neither atos or addr2line has been found"
                         " to resolve symbol names");
#endif //ATOS check

#else
    throw illformederror("No dlfcn.h utility exists for resolving symbol names.  Cannot use graphviz call tracer.");
#endif //DLFCN check

    fxns_[addr] = fxn;
}
#else
void
GraphViz::get_info(
    void* addr,
    std::string& fxn
)
{
    sst_throw(unimplementederror, "SST/macro has not been compiled with graphviz support");
}

#endif

GraphViz::Trace::Trace(const GraphViz::ptr& parent, const std::string& fxn)
    : parent_(parent), self_(0), fxn_(fxn)
{

}

GraphViz::Trace::ptr
GraphViz::get_trace(const std::string& fxn)
{
    return traces_[fxn];
}

GraphViz::ptr
GraphViz::construct()
{
    return new GraphViz;
}

void
GraphViz::simulation_finished(timestamp t)
{
    check_open();
    myfile << "events: Instructions\n\n";

    std::map<std::string,Trace::ptr>::const_iterator it = traces_.begin(), end = traces_.end();
    for (; it != end; ++it)
    {
        myfile << it->second->summary();
        myfile << "\n";
    }
    myfile.close();
}

GraphViz::GraphViz()
    : stat_logger("<stats> callgraph | graphviz", "callgrind.out")
{
}

void
GraphViz::add_call(
    int ncalls,
    long count,
    void* fxn,
    void* callfxn
)
{
    std::string fxn_name; get_info(fxn, fxn_name);
    std::string callfxn_name; get_info(callfxn, callfxn_name);
    Trace::ptr tr = traces_[fxn_name];
    if (!tr)
    {
        tr = Trace::construct(GraphViz::ptr(this), fxn_name);
        traces_[fxn_name] = tr;
    }
    tr->add_call(callfxn_name, ncalls, count);
}

void
GraphViz::trace(long count)
{
#if !HAVE_GRAPHVIZ
    return; //nothing to do here
#endif

    if (!is_active())
        return;

    void** stack = 0;
    if (trace_pool_.empty()){
        stack = new graphviz_trace;
    }
    else{
        stack = trace_pool_.front();
        trace_pool_.pop_front();
    }


#if HAVE_EXECINFO_H
    int nfxn = backtrace(stack, BACKTRACE_NFXN);
#else
    int nfxn = 0;
    throw illformederror("No execinfo.h utility exists for collecting backtraces.  Cannot use graphviz call tracer.");
#endif

    /** ignore the first and last two calls in the stack -
        these are just block and trace and main and run_thread */

    /** see how much of the backtrace is new.  we don't
        want to double count the number of calls */
    int start = nfxn - 3;
    int stackpos = start;

    thread::ptr thr = OS_API::get_current_thread();
    long thrid = thr->get_id();
    void** prev_stack = thr->get_backtrace();
    int prev_stackpos = prev_stack ? thr->get_backtrace_size() - 3 : 0;

    while (stackpos > 2 && prev_stackpos > 2)
    {
        void* fxn = stack[stackpos];
        void* prev_fxn = prev_stack[prev_stackpos];
        if (fxn == prev_fxn)
        {
            void* callfxn = stack[stackpos - 1];
            add_call(0, count, fxn, callfxn);
            --stackpos;
            --prev_stackpos;
        }
        else
        {
            break;
        }
    }

    while (stackpos > 2)
    {
        void* fxn = stack[stackpos];
        void* callfxn = stack[stackpos - 1];
        add_call(1, count, fxn, callfxn);
        --stackpos;
    }

    if (nfxn >= 2)
    {
        void* addr = stack[2];
        std::string fxn; get_info(addr, fxn);
        Trace::ptr tr = traces_[fxn];
        if (!tr)
        {
            tr = Trace::construct(GraphViz::ptr(this), fxn);
            traces_[fxn] = tr;
        }
        tr->add_self(count);
    }

    thr->set_backtrace(nfxn, stack);
    if (prev_stack)
    {
        trace_pool_.push_back(prev_stack);
    }

}

    }
}


