/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_PROCESS_operatingsystem_H_INCLUDED
#define SSTMAC_SOFTWARE_PROCESS_operatingsystem_H_INCLUDED

#include <sstmac/software/process/os_api.h>
//#include <sstmac/software/threading/context.h>
#include <sstmac/software/threading/threading_interface.h>
#include <sstmac/software/threading/stackalloc.h>
#include <sstmac/software/ami/ami.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/software/ami/hardware_descriptor.h>
#include <sstmac/software/libraries/library.h>
#include <sstmac/software/process/thread.h>
#include <sstmac/software/process/app.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/software/services/launch/launchinfo.h>

#include <stack>
#include <list>
#include <string>

namespace sstmac
{
  namespace hw {
    class computescheduler;
  }
  namespace sw
  {

    class environment;
    class service;
    class GraphViz;

    class operatingsystem : public OS_API
    {

      friend class service;
      friend class thread;
      friend class hw::computescheduler;


    protected:
      operatingsystem(hardware_descriptor::ptr &hw);

      /// This is how we keep our chucks of stack.
      static stackalloc stackalloc_;

      static std::list<boost::intrusive_ptr<thread> > to_delete_;

      std::map<taskid, long> task_to_thread_;

      /// The currently active position in the list.
      //  threadcontext_t *current_;

      static debug dbg_;

      static boost::intrusive_ptr<GraphViz> call_graph_;

      bool restarting_;

      int ncores_;

      int current_thread_id_;

      boost::intrusive_ptr<hw::computescheduler> thread_scheduler_;

      /// The caller context (main DES thread).  We go back
      /// to this context on every context switch.
      threading_interface *des_context_;

      void
      add_thread(const thread::ptr &t);

      void
      switch_to_thread(thread_data_t tothread);

    public:
      typedef boost::intrusive_ptr<operatingsystem> ptr;

      virtual
      ~operatingsystem();

      static operatingsystem::ptr
      construct(hardware_descriptor::ptr &hw);

      // --------------------------------------//

    public:

      void
      cleanup();

      void
      restarting(bool res)
      {
        restarting_ = res;
      }

      // ---- OS_API functions ------//
      // -- these are exposed to libraries --
      virtual void execute_kernel(AMI::AMI_COMP_FUNC func, const sst_message::ptr& data);
      virtual ptr_type::ptr execute_kernel(AMI::AMI_COMP_FUNC func);
      virtual void execute_kernel(AMI::AMI_COMM_FUNC func, const sst_message::ptr& data);
      virtual ptr_type::ptr execute_kernel(AMI::AMI_COMM_FUNC func);
      virtual void execute_kernel(AMI::AMI_HW_FUNC func, const sst_message::ptr& data);
      virtual ptr_type::ptr execute_kernel(AMI::AMI_HW_FUNC func);
      virtual void execute_kernel(AMI::AMI_QUERY_FUNC func, const sst_message::ptr& data);
      virtual ptr_type::ptr execute_kernel(AMI::AMI_QUERY_FUNC func);
      virtual void execute_kernel(AMI::AMI_DISP_FUNC func, const sst_message::ptr& data);
      virtual ptr_type::ptr execute_kernel(AMI::AMI_DISP_FUNC func);

      virtual bool kernel_supported(AMI::AMI_COMP_FUNC);
      virtual bool kernel_supported(AMI::AMI_COMM_FUNC);
      virtual bool kernel_supported(AMI::AMI_DISP_FUNC);
      virtual bool kernel_supported(AMI::AMI_QUERY_FUNC);
      virtual bool kernel_supported(AMI::AMI_HW_FUNC);

      static void stack_check();

      virtual timestamp
      block(const key::ptr &req);
      virtual timestamp
      unblock(const key::ptr &req);
      virtual void
      start_thread(const thread::ptr &t);
      //virtual void
      //join_thread(const thread::ptr &t, threadcontext_t con);
      virtual void
      complete_thread(bool succ);

      virtual void
      register_lib(const library::ptr &lib);

      virtual void
      finish();

      virtual void
      unregister_lib(const library::ptr &lib);

      virtual void
      register_ser(const boost::intrusive_ptr<service> &lib);

      virtual library::ptr
      get_lib(std::string name);

      virtual timestamp
      now() const;

      virtual nodeaddress::ptr
      get_task_addr(softwareid sid) const;

      virtual environment::const_ptr
      get_env(appid aid) const;

      void set_eventmanager(const boost::intrusive_ptr<eventmanager>& man);

      // ---------------------

      // ---- functions which can be called by services ------- //
      virtual nodeaddress::ptr
      get_my_addr() const;

      long get_task_threadid(const taskid& id) const;

      // --- functions here to be called by node / core / hardware --- //

      void
      setAMI(AMI::AMI_CAT cat, AMI* a);

      void
      add_application(const app::ptr& a, launchinfo::ptr &linfo);

      void
      add_task(const taskid& tid);

      void
      set_environment(const environment::const_ptr &e, appid aid)
      {
        SSTMAC_DEBUG << "OS: setting environment for app " << aid.to_string()
              << "\n";
        env_[aid] = e;
      }

      void
      init_scheduler(const boost::intrusive_ptr<hw::computescheduler>& sched);

      void
      start_apps();

      void
      handle_message(const sst_message::ptr &msg);

      void
      apps_kill();

      void
      kill_app(app::ptr app);

      std::list<app::ptr>
      get_app_ptrs(appid aid);

      app::ptr
      get_app_ptr(softwareid sid);

      // -------------------------------------- //

      virtual thread_data_t
      get_current_context() const
      {
        return threadstack_.top();
      }



      virtual long
      get_current_threadid() const;

      AMI*
      getAMI(AMI::AMI_CAT cat) const;

    protected:
      // scheduler scheduler_;
      ///hardware_descriptor::ptr hw_desc_;

      std::map<sstmac::sw::appid, sstmac::sw::environment::const_ptr> env_;

      std::map<AMI::AMI_CAT, AMI*> amis_;
      std::map<std::string, library::ptr> libs_;

      std::list<thread::ptr> threads_;

      std::stack<thread_data_t> threadstack_;

      std::deque<sst_message::ptr> pending_messages_;

      // threading_interface *threader_;
    };

  }
} //end of namespace sstmac
#endif
