/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/threading/stackalloc.h>
#include <sstmac/software/threading/stackalloc-chunk.h>

#include <sstmac/common/errors.h>
#include <unistd.h>


namespace sstmac { namespace sw {
  //
  // Build.
  //
  stackalloc::stackalloc(size_t stacksize, size_t alloc_unit, bool use_mprot) :
    suggested_chunk_(alloc_unit), stacksize_(stacksize), use_mprot_(use_mprot)
  {
    size_t rem = stacksize_ % sysconf(_SC_PAGESIZE);
    if(rem) {
      stacksize_ += (sysconf(_SC_PAGESIZE) - rem);
    }
  }

  //
  // Goodbye.
  stackalloc::~stackalloc() {
      chunk_vec_t::iterator it = chunks_.begin(), end = chunks_.end();
      for ( ; it != end; ++it){
        chunk* ch = *it;
        delete ch;
      }
  }

  //
  // Access the allocation size.
  //
  size_t stackalloc::stacksize() const {
    return stacksize_;
  }

  //
  // Get a stack memory region.
  //
  void* stackalloc::alloc() {
    if(available_.empty()) {
      //std::cerr << "stackalloc::alloc:  Allocating a new chunk\n";
      // grab a new chunk.
      void *buf = 0;
      while(! buf) {
        if(! chunks_.empty()) {
          buf = chunks_.back()->get_next_stack();
        }
        if(! buf) {
          chunk* new_chunk = new chunk(stacksize_, suggested_chunk_, use_mprot_);
          chunks_.push_back(new_chunk);
        }
      }
      //std::cerr << "stackalloc::alloc:  buf==" << buf << "\n";
      return buf;
    }
    else {
      void *buf = available_.back();
      available_.pop_back();
      return buf;
    }
  }

  //
  // Return the given memory region.
  //
  void stackalloc::free(void* buf) {
    available_.push_back(buf);
  }



} } // end of namespace sstmac
